{
  File Name: AudioIO.PAS  V 4.01
  Created: 5-Oct-96 by John Mertus on the IBM PC
  Revision #1: 5-Oct-22 by John Mertus
                                        -John Mertus

  Version 1.00 Initial Release as a unit
          2.00 Changed to a Component
          3.00 Elimated the window call.
          4.00 Moved to threads, this means the
               window component could be elimated,
               but that means a major rewrite so
               for now just threads.

}

{
  There are three Sound Components, the first is the base Component,
  TAudioIO.  This defines the sampling rates, buffers and some of the
  common events.

  The second component is AudioOut, which started just loops playing out
  buffers.

  The third component is AudioIN, which, when started, just loops filling
  buffer with digital data.

  See AudioIO.Hlp for detailed explaination.



}


{-----------------Unit-AudioOut-------------------John Mertus---Oct 96---}

                       Unit AudioIO;

{*************************************************************************}
                            Interface



uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Buttons, MMSystem;//, UAFDefs;

{ Could make this dynamic, but the effort doesn't seem worth it. }
Const
  MAXBUFFERS = 4;

Type

  {TBuffer Event is what is called when a buffer is need or is full }

  TBufferEvent = Function(Buffer : pChar; Var Size : Integer) : Boolean of Object;

{---------------------------TAudioIO Component-----------------------------}

  TAudioIO = class(TCustomControl)
  private
    { Private declarations }
    FBufferSize : Integer;            { Actual buffer used }
    FRequestedBufferSize : Integer;   { Buffer size requested }
    FNumBuffers          : Integer;
    FPaused              : Boolean;
    FKeepOpen            : Boolean;   { Keeps driver open }

    FWaveFmtEx         : TWaveFormatEx;
    FonOpen            : TNotifyEvent;
    FonClose           : TNotifyEvent;
    FWaveDevice        : DWord;
    hWaveHeader	       : Array [0..MAXBUFFERS-1] of THANDLE;
    hWaveBuffer        : Array [0..MAXBUFFERS-1] of THANDLE;
    BufIndex           : Integer;

    { Property Functions }
    Procedure SetNumBuffers(Value : Integer);
    Procedure SetBufferSize(Value : Integer);
    Procedure SetFrameRate(Value : Integer);
    Procedure SetStereo(Value : Boolean);
    Procedure SetBits(Value : Word);
    Function  GetFrameRate : Integer;
    Function  GetStereo : Boolean;
    Procedure MakeWaveFmtConsistent;

  protected
    { Protected declarations }
    WaveHdr            : Array [0..MAXBUFFERS-1] of PWAVEHDR;
    WaveBuffer         : Array [0..MAXBUFFERS-1] of lpstr;
    ContinueProcessing : Boolean;     { Set to TRUE to start FALSE to abort
                                         after filled buffers are done }
    AbortProcessing : Boolean;        { True to skip all buffers }
    Function InitWaveHeaders : Boolean;
    Function AllocPCMBuffers : Boolean;
    Function FreePCMBuffers : Boolean;
    Function AllocWaveHeaders : Boolean;
    Procedure FreeWaveHeaders;

  public
    { Public declarations }
    ErrorMessage   : String;
    FAudioActive    : Boolean;
    FilledBuffers,
    QueuedBuffers,
    ProcessedBuffers   : Integer;

    Constructor Create(AOwner: TComponent); Override;
    Destructor  Destroy;                    Override;
    Procedure   StopAtOnce;                 Virtual;
    Procedure   StopGracefully;             Virtual;
    Property    AudioActive : Boolean Read FAudioActive;
    Procedure   Paint;                      Override;

  published
    { Published declarations }
      Property BufferSize : Integer read FBufferSize write SetBufferSize Default 8192;
      Property NumBuffers : Integer read FNumBuffers write SetNumBuffers Default 4;
      Property FrameRate : Integer read GetFrameRate Write SetFrameRate Default 22055;
      Property Stereo : Boolean read GetStereo Write SetStereo Default False;
      Property Quantization : Word Read FWaveFmtEx.wBitsPerSample Write SetBits Default 16;
      Property WaveDevice : DWord Read FWaveDevice Write FWaveDevice Default WAVE_MAPPER;
      Property OnStart : TNotifyEvent Read FOnOpen Write FOnOpen;
      Property OnStop  : TNotifyEvent Read FOnClose Write FOnClose;
      Property KeepOpen : boolean     Read FKeepOpen Write FKeepOpen  Default False;
  end;


{---------------------------TAudioOut Component-----------------------------}


  TWaveOutThread = Class;
  TAudioOut = Class(TAudioIO)
  private
    { Private declarations }
    WaveDeviceOpen : Boolean;
    FOnFillBuffer  : TBufferEvent;

    Function  QueueBuffer : Boolean;
    Procedure SetPaused(Value : Boolean);
    Function  Setup : Boolean;  Virtual;
    Function  StartIt : Boolean;

    procedure BufferDone(var Msg: TMessage);
    procedure WaveOpen(var Msg: TMessage);
    procedure WaveClose;

  protected
    { Protected declarations }
    Procedure ShutdownWaveDevice;   Virtual;
    Procedure CloseWaveDevice;      Virtual;
    Function  ReadBuffer(Idx, N : Integer) : Boolean;  Virtual;

  public
    { Public declarations }
    WaveHandle     : HWaveOut;     { Waveform output handle }
    WaveOutThread  : TWaveOutThread;

    Destructor  Destroy;                          Override;
    Function  Start : Boolean;                    Virtual;
    Procedure StopAtOnce;                         Override;
    Procedure StopGracefully;                     Override;
    Function  ElapsedTime : Real;
    Function  ElapsedFrames : Integer;            Virtual;
    Procedure Paint;                              Override;

  published
    { Published declarations }
    Property Paused : Boolean Read FPaused Write SetPaused Default FALSE;
    Property OnFillBuffer : TBufferEvent Read FOnFillBuffer Write FOnFillBuffer;
  End;


{---------------------------TAudioIn Component-----------------------------}


  TWaveInThread = Class;
  TAudioIn = Class(TAudioIO)
  private
    { Private declarations }
    WaveDeviceOpen : Boolean;
    FOnBufferFilled  : TBufferEvent;
    FWaveReset       : Boolean;
    LastBufFrames    : Integer;

    Function  QueueBuffer : Boolean;
    Function  ProcessBuffer(B : lpstr; N : Integer) : Boolean;  Virtual;
    Procedure ShutdownWaveDevice;   Virtual;
    Procedure CloseWaveDevice;      Virtual;
    Procedure SetPaused(Value : Boolean);

    { These process the messages from the audio input system }
    procedure BufferFinished;
    procedure WaveOpenIn;
    procedure WaveCloseIn;

  protected
    { Protected declarations }

  public
    { Public declarations }
    WaveHandle        : HWaveOut;     { Waveform output handle }
    WaveInThread      : TWaveInThread;
    Destructor  Destroy;                        Override;
    Function  Start : Boolean;                  Virtual;
    Function  Setup : Boolean;                  Virtual;
    Function  StartIt : Boolean;
    Procedure StopAtOnce;                       Override;
    Procedure StopGracefully;                   Override;
    Function  ElapsedTime : Real;
    Function  ElapsedFrames : Integer;          Virtual;
    Procedure Paint;                            Override;

  published
    { Published declarations }
    Property Paused : Boolean Read FPaused Write SetPaused Default FALSE;
    Property OnBufferFilled : TBufferEvent Read FOnBufferFilled Write FOnBufferFilled;
  End;

  { This is the thread that receives the messages for output}
  TWaveOutThread = Class(TThread)
    procedure Execute; override;

    Private
      FAudioOut : TAudioOut;
      Procedure  WaveClose;
      Procedure  BufferDone;

    Public
      Constructor Create(AudioOut: TAudioOut);
  End;

  { This is the thread that receives the messages for Input}
  TWaveInThread = Class(TThread)
    procedure Execute; override;

    Private
      FAudioIn : TAudioIn;
      Procedure  WaveClose;
      Procedure  BufferDone;

    Public
      Constructor Create(AudioIn: TAudioIn);
  End;

var
  iVar : Integer;

procedure Register;

{*************************************************************************}

         implementation

//{$R *.res}

{---------------TWaveOutGetErrorText------------John Mertus Oct 96---}

   Function TWaveOutGetErrorText(iErr : Integer) : String;

{  This just gets the error text assocated with the output error ierr.  }
{									}
{**********************************************************************}
Var
  ErrorMsgC   : Array [0..255] of Char;

BEGIN
  waveOutGetErrorText(iErr,ErrorMsgC,Sizeof(ErrorMsgC));
  Result := StrPas(ErrorMsgC);
END;

{---------------TWaveInGetErrorText------------John Mertus Oct 96---}

   Function TWaveInGetErrorText(iErr : Integer) : String;

{  This just gets the error text assocated with the output error ierr.  }
{									}
{**********************************************************************}
Var
  ErrorMsgC   : Array [0..255] of Char;

BEGIN
  waveInGetErrorText(iErr,ErrorMsgC,Sizeof(ErrorMsgC));
  Result := StrPas(ErrorMsgC);
END;

procedure Register;
begin
  RegisterComponents('manip', [TAudioOut, TAudioIn]);
end;

{---------------SetBufferSize-------------------John Mertus Oct 96---}

   Procedure TAudioIO.SetBufferSize(Value : Integer);

{  This just set the buffersize, making sure it is too small.           }
{									}
{**********************************************************************}
BEGIN
   If (Value < 512) Then Value := 512;
  { make the wave buffer size a multiple of the block align... }
   FRequestedBufferSize := Value;
   MakeWaveFmtConsistent;
   FreePCMBuffers;
   AllocPCMBuffers;
END;

{---------------SetNumBuffers-------------------John Mertus Oct 96---}

   Procedure TAudioIO.SetNumBuffers(Value : Integer);

{  This just set the numbers of buffers making sure it is between       }
{ and MaxNumberBuffers                                                  }
{									}
{**********************************************************************}
BEGIN
   If (Value < 2) Then Value := 2;
   If (Value > MAXBUFFERS) Then Value := MAXBUFFERS;
   FNumBuffers := Value;
END;

{---------------SetStereo-----------------------John Mertus Oct 96---}

   Procedure TAudioIO.SetStereo(Value : Boolean);

{  This just set the numbers of channels, True 2, false 1.              }
{									}
{**********************************************************************}
BEGIN
  If Value Then
     FWaveFmtEx.nChannels := 2
  Else
     FWaveFmtEx.nChannels := 1;
   MakeWaveFmtConsistent;
END;

{---------------SetBits-------------------------John Mertus Oct 96---}

   Procedure TAudioIO.SetBits(Value : Word);

{  This just set the numbers of buffers making sure it is between       }
{ and MaxNumberBuffers                                                  }
{									}
{**********************************************************************}
BEGIN
   If (Value < 8) Then Value := 8;
   If (Value > 8) and (Value <= 16) Then Value := 16;
   If (Value > 16) Then Value := 24;
   FWaveFmtEx.wBitsPerSample := Value;
   MakeWaveFmtConsistent;
END;

{---------------SetFrameRate--------------------John Mertus Oct 96---}

   Procedure TAudioIO.SetFrameRate(Value : Integer);

{  This just set the frame rate for sampling.                           }
{									}
{**********************************************************************}
BEGIN
   FWaveFmtEx.nSamplesPerSec := Value;
   MakeWaveFmtConsistent;
END;

{---------------GetFrameRate--------------------John Mertus Oct 96---}

   Function TAudioIO.GetFrameRate : Integer;

{  This just returns the framerate for the current header.              }
{									}
{**********************************************************************}
BEGIN
  Result := FWaveFmtEx.nSamplesPerSec;
END;

{---------------GetStereo-----------------------John Mertus Oct 96---}

   Function TAudioIO.GetStereo : Boolean;

{  This just returns the True if stereo, e.g. 2 channels                }
{									}
{**********************************************************************}
BEGIN
  Result := (FWaveFmtEx.nChannels = 2);
END;


{-----------------Create------------------------John Mertus Oct 96---}

   Constructor TAudioIO.Create(AOwner: TComponent);

{  This just set the numbers of buffers making sure it is between       }
{ and MaxNumberBuffers                                                  }
{									}
{**********************************************************************}
Var
  i : Integer;


BEGIN
   Height := 28;
   Width := 28;
   Inherited Create(AOwner);
   FNumBuffers := 4;
   FRequestedBufferSize := 8192;
   FAudioActive := FALSE;
   FPaused := FALSE;
   FWaveDevice := WAVE_MAPPER;
   ErrorMessage := '';
   KeepOpen := False;

  { Set the indendent sampling rates }
   FWaveFmtEx.wFormatTag := WAVE_FORMAT_PCM;
   FWaveFmtEx.wBitsPerSample := 16;
   FWaveFmtEx.nchannels := 1;
   FWaveFmtEx.nSamplesPerSec := 22050;
   MakeWaveFmtConsistent;

  { Now make sure we know buffers are not allocated }
   For i := 0 to MAXBUFFERS-1 Do WaveBuffer[i] := Nil;

   AllocWaveHeaders;
   AllocPCMBuffers;
END;

{-----------------Destroy-----------------------John Mertus Oct 96---}

   Destructor TAudioIO.Destroy;

{  This cleans up the buffers.                                          }
{									}
{**********************************************************************}
BEGIN
  FreePCMBuffers;
  FreeWaveHeaders;
  Inherited Destroy;
END;

{-----------------MakeWaveFmtConsistent---------John Mertus Oct 96---}

   Procedure TAudioIO.MakeWaveFmtConsistent;

{  This just trys to find the correct avgbytes and blockalign that      }
{ one needs to use for the format.  I DO NOT UNDERSTAND WHY MICROSOFT   }
{ did this.                                                             }
{									}
{**********************************************************************}
BEGIN
   With FWaveFmtEx Do
     Begin
        nBlockAlign := (wBitsPerSample div 8)*nchannels;
        nAvgBytesPerSec := nSamplesPerSec*nBlockAlign;
     End;

   FBufferSize := FRequestedBufferSize - (FRequestedBufferSize mod FWaveFmtEx.nBlockAlign);
END;

{-------------InitWaveHeaders----------------John Mertus---14-June--97--}

   Function TAudioIO.InitWaveHeaders : Boolean;

{ This just initalizes the waveform headers, no memory allocated        }
{									}
{**********************************************************************}
Var
  i : Integer;

BEGIN
  { This should not be necessary, but to be safe... }
  MakeWaveFmtConsistent;

  { Set the wave headers }
  For i := 0 to FNumBuffers-1 Do
    With WaveHdr[i]^ Do
      Begin
        lpData := WaveBuffer[i];         // address of the waveform buffer
        dwBufferLength := FBufferSize;   // length, in bytes, of the buffer
        dwBytesRecorded := 0;            // see below
        dwUser := 0;                     // 32 bits of user data
        dwFlags := 0;                    // see below
        dwLoops := 0;                    // see below
        lpNext := Nil;                   // reserved; must be zero
        reserved := 0;                   // reserved; must be zero
      End;

  InitWaveHeaders := TRUE;
END;


{-------------AllocPCMBuffers----------------John Mertus---14-June--97--}

   Function TAudioIO.AllocPCMBuffers : Boolean;

{ Allocate and lock the waveform memory.                                }
{									}
{***********************************************************************}
Var
  i : Integer;

BEGIN
  For i := 0 to fNumBuffers-1 Do
    begin
      hWaveBuffer[i] := GlobalAlloc( GMEM_MOVEABLE or GMEM_SHARE, fBufferSize );
      If (hWaveBuffer[i] = 0) Then
	begin
          FreePCMBuffers;
          ErrorMessage := 'Error allocating wave buffer memory';
          AllocPCMBuffers := False;
          Exit;
	end;

      WaveBuffer[i] := GlobalLock(hWaveBuffer[i]);
      If (WaveBuffer[i] = Nil) Then
	begin
          FreePCMBuffers;
          ErrorMessage := 'Error Locking wave buffer memory';
          AllocPCMBuffers := False;
          Exit;
	end;
      WaveHdr[i].lpData := WaveBuffer[i];
    End;

  AllocPCMBuffers := TRUE;
END;

{--------------FreePCMBuffers----------------John Mertus---14-June--97--}

   Function TAudioIO.FreePCMBuffers : Boolean;

{ Free up the meomry AllocPCMBuffers used.                              }
{									}
{***********************************************************************}
Var
  i : Integer;

BEGIN

  Result := FALSE;

  For i := 0 to MaxBuffers-1 Do
    begin
      If (hWaveBuffer[i] <> 0) Then
        Begin
  	  GlobalUnlock(hWaveBuffer[i] );
	  GlobalFree(hWaveBuffer[i] );
          hWaveBuffer[i] := 0;
          WaveBuffer[i] := Nil;
          Result := TRUE;
        End;
    end;

END;
{-------------AllocWaveHeaders---------------John Mertus---14-June--97--}

   Function TAudioIO.AllocWaveHeaders : Boolean;

{ Allocate and lock header memory                                       }
{									}
{***********************************************************************}
Var
  i : Integer;

BEGIN
  For i := 0 to MAXBUFFERS-1 Do
    begin
      hwaveheader[i] := GlobalAlloc( GMEM_MOVEABLE or GMEM_SHARE or GMEM_ZEROINIT, sizeof(TWAVEHDR));
      if (hwaveheader[i] = 0) Then
        begin
          FreeWaveHeaders;
          ErrorMessage := 'Error allocating wave header memory';
          AllocWaveHeaders := FALSE;
          Exit;
        end;

      WaveHdr[i] := GlobalLock (hwaveheader[i]);
      If (WaveHdr[i] = Nil ) Then
        begin
          FreeWaveHeaders;
          ErrorMessage := 'Could not lock header memory for recording';
          AllocWaveHeaders := FALSE;
          Exit;
        end;

    End;

  AllocWaveHeaders := TRUE;
END;

{---------------FreeWaveHeaders---------------John Mertus---14-June--97--}

   Procedure TAudioIO.FreeWaveHeaders;

{ Just free up the memory AllocWaveHeaders allocated.                   }
{									}
{***********************************************************************}
Var
  i : Integer;

BEGIN
  For i := 0 to MAXBUFFERS-1 Do
    begin
      If (hWaveHeader[i] <> 0) Then
        Begin
          GlobalUnlock(hwaveheader[i]);
          GlobalFree(hwaveheader[i]);
          hWaveHeader[i] := 0;
          WaveHdr[i] := Nil;
        End
    end;
END;

{--------------------StopAtOnce-------------John Mertus---14-June--97--}

   Procedure TAudioIO.StopAtOnce;

{ Write the buffer to the wave device and toggel buffer index.          }
{									}
{**********************************************************************}
BEGIN
   FAudioActive := False;
   ContinueProcessing := FALSE;
   AbortProcessing := TRUE;
END;

{--------------------StopGracefully---------John Mertus---14-June--97--}

   Procedure TAudioIO.StopGracefully;

{ Write the buffer to the wave device and toggel buffer index.          }
{									}
{**********************************************************************}
BEGIN
  ContinueProcessing := FALSE;
END;

{-----------------Destroy-----------------------John Mertus Oct 96---}

   Destructor TAudioOut.Destroy;

{  This cleans up the buffers.                                          }
{									}
{**********************************************************************}
BEGIN
  StopAtOnce;
  CloseWaveDevice;
  Inherited Destroy;
END;

{-----------------ElapsedTime----------------John Mertus---14-June--97--}

   Function TAudioOut.ElapsedTime : Real;

{ This function returns the time since start of playout.                }
{									}
{**********************************************************************}
BEGIN
  If (AudioActive) Then
    Result := ElapsedFrames/FrameRate
  Else
    Result := 0;
END;

{-----------------ElapsedFrames--------------John Mertus---14-June--97--}

   Function TAudioOut.ElapsedFrames : Integer;

{ This function returns the frames played since start of playout.       }
{									}
{**********************************************************************}
Var
  pmmt  : TMMTime;

BEGIN
  If (AudioActive) Then
    Begin
      pmmt.wType := TIME_SAMPLES;
      If (waveOutGetPosition(WaveHandle, @pmmt, Sizeof(TMMTime)) <> 0) Then
         Result := 0
      Else
         Result := pmmt.Sample;
    End
  Else
    Result := 0;
END;

{---------------SetPaused-----------------------John Mertus Oct 96---}

   Procedure TAudioOut.SetPaused(Value : Boolean);

{  This pauses or restarts the output.                                  }
{									}
{**********************************************************************}
BEGIN
  FPaused := Value;
  If (Not AudioActive) Then Exit;
  If FPaused Then
    WaveOutPause(WaveHandle)
  Else
    WaveOutReStart(WaveHandle);
END;

{-------------ShutdownWaveDevice----------------John Mertus---14-June--97--}

   Procedure TAudioOut.ShutdownWaveDevice;

{ This just cleans up the shutdown                                      }
{									}
{**********************************************************************}
Var
  i : Integer;

BEGIN
  If (Not WaveDeviceOpen) Then Exit;

{ Destroy the thread }
  If (Assigned(WaveOutThread)) Then
    Begin
      If (Not KeepOpen) Then
        WaveOutThread.Terminate
      Else
        WaveClose;
    End;
    
  WaveOutThread := Nil;

  WaveOutReset(WaveHandle);
  FAudioActive := FALSE;
  Paused := FALSE;
  For i := 0 to FNumBuffers-1 Do
    waveOutUnprepareHeader( WaveHandle, WaveHdr[i], sizeof(TWAVEHDR));

{ close the device }
  If (Not KeepOpen) Then CloseWaveDevice;

END;

{---------------CloseWaveDevice----------------John Mertus---14-June--97--}

   Procedure TAudioOut.CloseWaveDevice;

{ Closes the wave output device.                                        }
{       									}
{**********************************************************************}

BEGIN
   If (Not WaveDeviceOpen) Then Exit;
   WaveDeviceOpen := FALSE;
   WaveOutReset(WaveHandle);
   waveOutClose(WaveHandle);    // This will terminate the thread
END;

{----------------------------TWaveOutThread-----------------------------}

{---------------Create-----------------------John Mertus---14-March--99-}

  Constructor TWaveOutThread.Create(AudioOut : TAudioOut);

{ This is run as a thread and processes all the messages for the        }
{ audio.                                                                }
{									}
{**********************************************************************}

BEGIN
  inherited Create(False);
  FAudioOut := AudioOut;
  FreeOnTerminate := TRUE;
  Priority := tpTimeCritical;
END;

{---------------BufferDone------------------John Mertus---14-March--99-}

  Procedure TWaveOutThread.BufferDone;

{ The only purpose of this routine is to synchronize the output thread  }
{ with the main form on a buffer compleation.                           }
{									}
{**********************************************************************}
  Var
     msg        : TMessage;
  Begin
    msg.Msg := WOM_DONE;
    FAudioOut.BufferDone(msg);
  End;

{---------------WaveClose--------------------John Mertus---14-March--99-}

  Procedure TWaveOutThread.WaveClose;

{ The only purpose of this routine is to synchronize the output thread  }
{ with the main form on an close.                                       }
{									}
{**********************************************************************}
  Begin
    FAudioOut.WaveClose;
  End;

{---------------Execute----------------------John Mertus---14-March--99-}

  Procedure TWaveOutThread.Execute;

{ This is run as a thread and processes all the messages for the        }
{ audio.                                                                }
{									}
{**********************************************************************}

Var
  m          : TMsg;

BEGIN
 { Main message loop }
  While not Terminated do
    Begin
      If GetMessage(m, 0 , 0, 0) Then
        Begin
          Case m.message of
          WOM_OPEN:
            Begin
//  Removed and done in the startit code.
//              Synchronize(WaveOpen);
            End;

          WOM_DONE:
            Begin
              Synchronize(BufferDone);
            End;

          WOM_CLOSE:
            Begin
              Terminate;
            End;

          WM_USER:
            Begin
               // Ignore user messages
            End;
          Else
            MessageDlg(Format('Unexpected Out Thread Message %x',[m.message]),mtInformation, [mbOk], 0 );
          End;
        End;
    End;
    Synchronize(WaveClose);
END;

{-----------------------------TWaveInThread-----------------------------}

{---------------Create-----------------------John Mertus---14-March--99-}

  Constructor TWaveInThread.Create(AudioIn : TAudioIn);

{ This is run as a thread and processes all the messages for the        }
{ audio.                                                                }
{									}
{**********************************************************************}

BEGIN
  Inherited Create(False);
  FAudioIn := AudioIn;
  FreeOnTerminate := TRUE;
//  Priority := tpTimeCritical;
END;


{---------------BufferDone------------------John Mertus---14-March--99-}

  Procedure TWaveInThread.BufferDone;

{ The only purpose of this routine is to synchronize the output thread  }
{ with the main form on a buffer compleation.                           }
{									}
{**********************************************************************}
  Begin
    FAudioIn.BufferFinished;
  End;

{---------------WaveClose--------------------John Mertus---14-March--99-}

  Procedure TWaveInThread.WaveClose;

{ The only purpose of this routine is to synchronize the output thread  }
{ with the main form on an close.                                       }
{									}
{**********************************************************************}
  Begin
    FAudioIn.WaveCloseIn;
  End;

{---------------SetPaused-----------------------John Mertus Oct 96---}

   Procedure TAudioIn.SetPaused(Value : Boolean);

{  This pauses or restarts the output.                                  }
{									}
{**********************************************************************}
Var
  iErr : Integer;

BEGIN
  // Do nothing if the value is the same
  If (FPaused = Value) Then Exit;
  FPaused := Value;

  // Do nothing if we are not active
  If (Not AudioActive) Then Exit;
  If FPaused Then
    Begin
      { stop playing }
      LastBufFrames := ElapsedFrames;
      iErr := waveInReset(WaveHandle);
      If (iErr <> 0) Then ShowMessage(TWaveInGetErrorText(iErr));
           WaveInReset(WaveHandle)
    End
  Else
    Begin
      ProcessedBuffers := 0;
      iErr := WaveInStart(WaveHandle);
      If (iErr <> 0) Then
        begin
          ShutdownWaveDevice;
          ErrorMessage := 'Error starting wave record: ' + TWaveInGetErrorText(iErr);
          Exit;
        end;
    End;
END;

{---------------Execute----------------------John Mertus---14-March--99-}

  Procedure TWaveInThread.Execute;

{ This is run as a thread and processes all the messages for the        }
{ audio.                                                                }
{									}
{**********************************************************************}

Var
  m          : TMsg;

BEGIN
 { Main message loop }
  While not Terminated do
    Begin
      If GetMessage(m, 0 , 0, 0) Then
        Begin
          Case m.message of
          MM_WIM_OPEN:
            Begin
//  Removed and done in the startit code.
//              Synchronize(WaveOpen);
            End;

          MM_WIM_DATA:
            Begin
              Synchronize(BufferDone);
            End;

          MM_WIM_CLOSE:
            Begin
              Terminate;
            End;
          End;
        End;
    End;
    // We are done, say we are closed
    Synchronize(WaveClose);
END;

{-------------SetupOutput--------------------John Mertus---14-June--97--}

   Function TAudioOut.Setup : Boolean;

{ This function just sets up the board for output.                      }
{									}
{**********************************************************************}
Var
  iErr       : Integer;
  i          : Integer;

BEGIN

  { if the device is still open, return error }
  If AudioActive Then
    Begin
      ErrorMessage := 'Wave output device is already active';
      Result := FALSE;
      Exit;
    End;

  BufIndex := 0;

  { If not open, open the device for playout.  Keepopen will control if the
    driver is kept opened }
  If (Not WaveDeviceOpen) Then
    Begin
      { Create a thread for playout }
      WaveOutThread := TWaveOutThread.Create(Self);
      If (Not Assigned(WaveOutThread)) Then
        Begin
          ErrorMessage := 'Could not start thread';
          Result := FALSE;
          Exit;
        End;

      WaveHandle := 0;
      iErr := WaveOutOpen(@WaveHandle, FWaveDevice, @FWaveFmtEx, WaveOutThread.ThreadID,
                         0, CALLBACK_THREAD );

      If (iErr <> 0) Then
        Begin
          ErrorMessage := TWaveOutGetErrorText(iErr);
          Result := FALSE;
          Exit;
        End;
    End;

  WaveDeviceOpen := TRUE;

  { Setup the buffers and headers }
  If (Not InitWaveHeaders) Then
    Begin
      Result := FALSE;
      Exit;
    End;

  { Now Prepare the buffers for output }
  For i := 0 to FNumBuffers-1 Do
    Begin
      iErr := WaveOutPrepareHeader(WaveHandle, WaveHdr[i], sizeof(TWAVEHDR));
      If (iErr <> 0) Then
        Begin
          ErrorMessage := TWaveOutGetErrorText(iErr);
          ShutdownWaveDevice;
          Result := FALSE;
          Exit;
        End;
    End;

  { Read in the buffers }
  QueuedBuffers := 0;
  ProcessedBuffers := 0;
  FilledBuffers := 0;
  ContinueProcessing := TRUE;
  AbortProcessing := FALSE;
  FAudioActive := TRUE;

  If (Not ReadBuffer(0, FBufferSize)) Then
    Begin
      ShutdownWaveDevice;
      ErrorMessage := 'There must be at least one filled buffer';
      Result := FALSE;
      Exit;
    End;

  For i := 1 to FNumBuffers - 1 Do ReadBuffer(i, FBufferSize);

  Result := TRUE;
END;

{----------------QueueBuffer----------------John Mertus---14-June--97--}

   Function TAudioOut.QueueBuffer : Boolean;

{ Write the buffer to the wave device and toggel buffer index.          }
{									}
{**********************************************************************}
Var
  iErr : Integer;

BEGIN
// If (WaveHdr[bufIndex].dwBufferLength = 0) Then Exit;

 { reset flags field (remove WHDR_DONE attribute) }
  WaveHdr[bufindex].dwFlags := WHDR_PREPARED;

 { now queue the buffer for output }
  iErr := waveOutWrite( WaveHandle, WaveHdr[bufindex], sizeof(TWAVEHDR));
  If (iErr <> 0) Then
    Begin
      ErrorMessage := TwaveOutGetErrorText(iErr);
      StopGracefully;
      Result := FALSE;
      Exit;
    End;

  { Advance index }
  bufindex := (bufindex+1) mod FNumBuffers;

  Result := TRUE;
END;

{-------------StartIt------------------------John Mertus---14-June--97--}

   Function TAudioOut.StartIt : Boolean;

{ This function just starts the waveform playing                        }
{									}
{**********************************************************************}
Var
  i : Integer;

BEGIN
  FAudioActive := TRUE;

  If (FPaused) Then WaveOutPause(WaveHandle);
  { Now we are ready to start the output }
  If (Not QueueBuffer) Then
    Begin
      ShutdownWaveDevice;
      Result := FALSE;
      Exit;
    End;

  // Move onOpen here so it works even in Keepopen is true
  If Assigned(FonOpen) Then FonOpen(Self);
  For i := 1 to FNumBuffers - 1 Do QueueBuffer;
  Result := TRUE;
END;

{-----------------Start----------------------John Mertus---14-June--97--}

   Function TAudioOut.Start : Boolean;

{ This function first sets up the output and then starts it.            }
{									}
{**********************************************************************}
BEGIN
  Result := Setup;
  If (Not Result) Then Exit;

  Result := StartIt;
  If (Not Result) Then Exit;
END;


{-------------ReadBuffer---------------------John Mertus---14-June--97--}

   Function  TAudioOut.ReadBuffer(Idx, N : Integer) : Boolean;

{ This is called whenver move buffer data is needed.                    }
{									}
{**********************************************************************}
Var
  NSize : Integer;

BEGIN
  { Do not call the read buffer routine if we want to stop }
  If (Not ContinueProcessing) Then
     Begin
        WaveHdr[idx].dwBufferLength := 0;
        Result := FALSE;
        Exit;
     End;

  { If assigned, process the buffer, Notice that the Size returned may not
    be the size sent,  so reset the output size }
//  ZeroMemory(WaveBuffer[idx], NSize);
  If Assigned(FOnFillBuffer) Then
    Begin
      NSize := N;
      Result := FOnFillBuffer(WaveBuffer[idx], NSize);
      WaveHdr[idx].dwBufferLength := NSize;
    End
  Else
    Begin
      Result := TRUE;
      WaveHdr[idx].dwBufferLength := N;
    End;

  { On a filled buffer, increment it }
  If (Result) Then FilledBuffers := FilledBuffers + 1;
  QueuedBuffers := FilledBuffers - ProcessedBuffers;
END;

{--------------------StopAtOnce-------------John Mertus---14-June--97--}

   Procedure TAudioOut.StopAtOnce;

{ Write the buffer to the wave device and toggel buffer index.          }
{									}
{**********************************************************************}
BEGIN

   FAudioActive := False;
   ContinueProcessing := FALSE;

  { close the device and unprepare the headers }
   ShutdownWaveDevice;
END;

{--------------------StopGracefully---------John Mertus---14-June--97--}

   Procedure TAudioOut.StopGracefully;

{ Write the buffer to the wave device and toggel buffer index.          }
{									}
{**********************************************************************}
BEGIN
  { if the device isn't open, just return }
  If (Not WaveDeviceOpen) Then Exit;
  ContinueProcessing := FALSE;
END;

{------------------BufferDone---------------John Mertus---14-June--97--}

   Procedure TAudioOut.BufferDone(Var Msg : TMessage);

{ This is called when a buffer si done playing                          }
{									}
{**********************************************************************}
BEGIN
    Begin
      ProcessedBuffers := ProcessedBuffers + 1;
      QueuedBuffers := FilledBuffers - ProcessedBuffers;
      FAudioActive := (QueuedBuffers > 0);
      If (ReadBuffer(BufIndex, FBufferSize)) Then QueueBuffer;

      If (Not AudioActive) then
        Begin
          ContinueProcessing := FALSE;
          ShutdownWaveDevice;
        End;

    End;
END;

{------------------WaveOpen-----------------John Mertus---14-June--97--}

   Procedure TAudioOut.WaveOpen(Var Msg : TMessage);

{ This is called at the termination of each buffer.                     }
{									}
{**********************************************************************}
BEGIN
  If Assigned(FonOpen) Then FonOpen(Self);
END;


{------------------WaveClose----------------John Mertus---14-June--97--}

   Procedure TAudioOut.WaveClose;

{ This is called at the termination of each buffer.                     }
{									}
{**********************************************************************}
BEGIN
  FAudioActive := FALSE;
  If Assigned(FonClose) Then FonClose(Self);
END;

{-----------------ElapsedTime----------------John Mertus---14-June--97--}

   Function TAudioIn.ElapsedTime : Real;

{ This function returns the time since start of playout.                }
{									}
{**********************************************************************}
BEGIN
  If (AudioActive) Then
    Result := ElapsedFrames/FrameRate
  Else
    Result := 0;
END;

{-------------ShutdownWaveDevice----------------John Mertus---14-June--97--}

   Procedure TAudioIn.ShutdownWaveDevice;

{ This cleans up the waveform input shutdown                            }
{									}
{**********************************************************************}
Var
  i : Integer;

BEGIN
{ Stop all the I/O }
  StopAtOnce;

{ Destroy the thread }
  If (Assigned(WaveInThread)) Then WaveInThread.Terminate;

{ unprepare the headers }
  For i := 0 to FNumBuffers-1 Do
    waveInUnprepareHeader( WaveHandle, WaveHdr[i], sizeof(TWAVEHDR));

{ close the device }
   If (Not KeepOpen) Then CloseWaveDevice;
   WaveInThread := Nil;
END;

{----------------CloseWaveDevice----------------John Mertus---14-June--97--}

   Procedure TAudioIn.CloseWaveDevice;

{ Closes the wave output device.                                        }
{									}
{**********************************************************************}
Var
  iErr : Integer;

BEGIN
    If (AudioActive) Then ShutDownWaveDevice;
    Application.ProcessMessages;
    FAudioActive := False;

    // See if we are open
    If (WaveHandle <> 0) Then
      Begin
        iErr := waveInClose(WaveHandle);
        If (iErr <> 0) Then
          Begin
            MessageDlg(TWaveInGetErrorText(iErr), mtError, [mbOk], 0);
          End;
      End;
    WaveDeviceOpen := FALSE;
    WaveHandle := 0;
END;

{-------------SetupOutput--------------------John Mertus---14-June--97--}

   Function TAudioIn.Setup : Boolean;

{ This function just sets up the board for output.                      }
{									}
{**********************************************************************}
Var
  iErr        : Integer;
  i : Integer;

BEGIN

  { if the device is still open, return error }
  If (AudioActive) Then
    Begin
      ErrorMessage := 'Wave Input device is already active';
      Result := FALSE;
      Exit;
    End;

  BufIndex := 0;


  { If Not Opened, then open the device for playout.  KeepOpen will control
    if the audio output driver is closed after open. }
  If (Not WaveDeviceOpen) Then
    Begin

      { Create a thread for Input }
      WaveInThread := TWaveInThread.Create(Self);
      If (Not Assigned(WaveInThread)) Then
        Begin
          ErrorMessage := 'Could not start thread';
          Result := FALSE;
          Exit;
        End;
        
      WaveHandle := 0;
      iErr := WaveInOpen(@WaveHandle, FWaveDevice, @FWaveFmtEx, WaveInThread.ThreadID,
                          0, CALLBACK_THREAD);

      If (iErr <> 0) Then
        Begin
          ErrorMessage := TWaveInGetErrorText(iErr);
          Result := FALSE;
          Exit;
        End;
    End;

  // Reset is used to find when we have a waveInReset.  THis means we
  // can process the last buffer.

  WaveDeviceOpen := TRUE;
  FWaveReset := FALSE;

  { Setup the buffers and headers }
  If (Not InitWaveHeaders) Then
    Begin
      Result := FALSE;
      Exit;
    End;

  { Now Prepare the buffers for output }

  For i := 0 to FNumBuffers-1 Do
    Begin
      iErr := WaveInPrepareHeader(WaveHandle, WaveHdr[i], sizeof(TWAVEHDR));
      If (iErr <> 0) Then
        Begin
          ErrorMessage := TWaveInGetErrorText(iErr);
          ShutdownWaveDevice;
          Result := FALSE;
          Exit;
        End;
    End;

  { Read in the buffers }
  QueuedBuffers := 0;
  ProcessedBuffers := 0;
  FilledBuffers := 0;
  ContinueProcessing := TRUE;
  AbortProcessing := FALSE;
  FAudioActive := TRUE;

  Result := TRUE;
END;

{----------------QueueBuffer----------------John Mertus---14-June--97--}

   Function TAudioIn.QueueBuffer : Boolean;

{ Write the buffer to the wave device and toggel buffer index.          }
{									}
{**********************************************************************}
Var
  iErr : Integer;

BEGIN
 { reset flags field (remove WHDR_DONE attribute) }
//  WaveHdr[bufindex].dwFlags := WHDR_PREPARED;

  iErr := waveInUnPrepareHeader(WaveHandle, WaveHdr[bufindex], sizeof(TWAVEHDR));
  If (iErr <> 0) Then
    Begin
      ErrorMessage := TWaveInGetErrorText(iErr);
      StopGracefully;
      Result := FALSE;
      Exit;
    End;

  iErr := waveInPrepareHeader(WaveHandle, WaveHdr[bufindex], sizeof(TWAVEHDR));
  If (iErr <> 0) Then
    Begin
      ErrorMessage := TWaveInGetErrorText(iErr);
      StopGracefully;
      Result := FALSE;
      Exit;
    End;

 { now queue the buffer for output }
  iErr := waveInAddBuffer( WaveHandle, WaveHdr[bufindex], sizeof(TWAVEHDR));
  If (iErr <> 0) Then
    Begin
      ErrorMessage := TWaveInGetErrorText(iErr);
      StopGracefully;
      Result := FALSE;
      Exit;
    End;

  { Advance index }
  bufindex := (bufindex+1) mod FNumBuffers;
  QueuedBuffers := QueuedBuffers + 1;

  Result := TRUE;
END;

{-------------StartIt------------------------John Mertus---14-June--97--}

   Function TAudioIn.StartIt : Boolean;

{ This function just starts the waveform playing                        }
{									}
{**********************************************************************}
Var
  i, iErr : Integer;

BEGIN
  QueuedBuffers := 0;
  ProcessedBuffers := 0;
  FilledBuffers := 0;
  ContinueProcessing := TRUE;
  AbortProcessing := FALSE;
  FAudioActive := TRUE;

  { Queue up all the buffers }
  For i := 0 to FNumBuffers - 1 Do
   If (Not QueueBuffer) Then
    Begin
      ShutdownWaveDevice;
      Result := FALSE;
      Exit;
    End;

  { start recording to first buffer }
  If (Not Paused) Then
    Begin
      iErr := WaveInStart(WaveHandle);
      If (iErr <> 0) Then
        begin
          ShutdownWaveDevice;
          ErrorMessage := 'Error starting wave record: ' + TWaveInGetErrorText(iErr);
          Result := FALSE;
          Exit;
        end;
    End;

  FAudioActive := TRUE;

  // Call the wave open command;
  WaveOpenIn;
  Result := TRUE;
END;

{-----------------Start----------------------John Mertus---14-June--97--}

   Function TAudioIn.Start : Boolean;

{ This function first sets up the output and then starts it.            }
{									}
{**********************************************************************}
BEGIN
  Result := Setup;
  If (Not Result) Then Exit;

  Result := StartIt;
  If (Not Result) Then Exit;
END;


{-----------ProcessBuffer---------------------John Mertus---14-June--97--}

   Function  TAudioIn.ProcessBuffer(B : lpstr; N : Integer) : Boolean;

{ This is called whenver move buffer data is needed.                    }
{									}
{**********************************************************************}

BEGIN
  { Do not call the read buffer routine if we want to stop }
  If (N <= 0) Then
     Begin
        Result := FALSE;
        Exit;
     End;

  { N can change, but we dont' care }
  If Assigned(FOnBufferFilled) Then
    Begin
      Result := FOnBufferFilled(B, N);
    End
  Else
    Result := TRUE;

  { On a filled buffer, increment it }
  If (Result) Then FilledBuffers := FilledBuffers + 1;
END;

{--------------------StopAtOnce-------------John Mertus---14-June--97--}

   Procedure TAudioIn.StopAtOnce;

{ Write the buffer to the wave device and toggel buffer index.          }
{									}
{**********************************************************************}
Var
  iErr : Integer;
  
BEGIN
  { if the device isn't open, just return }
   If (Not WaveDeviceOpen) Then Exit;

   FWaveReset := TRUE;
   ContinueProcessing := FALSE;

   { stop playing }
   LastBufFrames := ElapsedFrames;

   AbortProcessing := TRUE;
   iErr := waveInReset(WaveHandle);
   If (iErr <> 0) Then ShowMessage(TWaveInGetErrorText(iErr));
   FAudioActive := False;

  { We have to get it there }
   Application.ProcessMessages;
END;

{--------------------StopGracefully---------John Mertus---14-June--97--}

   Procedure TAudioIn.StopGracefully;

{ Write the buffer to the wave device and toggel buffer index.          }
{									}
{**********************************************************************}
BEGIN
  StopAtOnce;
  { if the device isn't open, just return }
//  If (Not WaveDeviceOpen) Then Exit;
//  ContinueProcessing := FALSE;
END;


{------------------BufferFinished-----------John Mertus---14-June--97--}

   Procedure TAudioIn.BufferFinished;

{ This is called when each buffer is filled.                            }
{									}
{**********************************************************************}
Var
  BSize : Integer;

BEGIN
// See if we have stopped, if so, finished all the samples in
// the queue

   If (AbortProcessing) Then
     Begin
       BSize := LastBufFrames*(Quantization div 8) - ProcessedBuffers*FBufferSize;
       If (BSize > FBufferSize) Then BSize := FBufferSize;
     End
   Else
      BSize := FBufferSize;

  // Do Not process any more buffers
   If (BSize <= 0) Then Exit;

   ProcessedBuffers := ProcessedBuffers + 1;
   QueuedBuffers := QueuedBuffers - 1;
   FAudioActive := (QueuedBuffers > 0) and (Not ABortProcessing);

   If (ProcessBuffer(WaveBuffer[BufIndex], BSize)) Then QueueBuffer;
   If (Not AudioActive) then
     Begin
       ContinueProcessing := FALSE;
       ShutdownWaveDevice;
     End;
END;

{------------------WaveOpenIn---------------John Mertus---14-June--97--}

   Procedure TAudioIn.WaveOpenIn;

{ This is called at the termination of each buffer.                     }
{									}
{**********************************************************************}
BEGIN
  If Assigned(FonOpen) Then FonOpen(Self);
END;


{------------------WaveCloseIn----------------John Mertus---14-June--97--}

   Procedure TAudioIn.WaveCloseIn;

{ This is called at the termination of each buffer.                     }
{									}
{**********************************************************************}
BEGIN
  If Assigned(FonClose) Then FonClose(Self);
END;

{-----------------Destroy-----------------------John Mertus Oct 96---}

   Destructor TAudioIn.Destroy;

{  This cleans up the buffers.                                          }
{									}
{**********************************************************************}
BEGIN
  CloseWaveDevice;
  Inherited Destroy;
END;

{-----------------ElapsedFrames--------------John Mertus---14-June--97--}

   Function TAudioIn.ElapsedFrames : Integer;

{ This function returns the frames played since start of playout.       }
{									}
{**********************************************************************}
Var
  pmmt  : TMMTime;

BEGIN
  If (AudioActive) Then
    Begin
      pmmt.wType := TIME_SAMPLES;
      If (waveInGetPosition(WaveHandle, @pmmt, Sizeof(TMMTime)) <> 0) Then
         Result := 0
      Else
         Result := pmmt.Sample;
    End
  Else
    Result := 0;
END;

{--------------------Paint-------------------John Mertus---14-June--97--}

    procedure TAudioIO.Paint;

{ This is used only at design time to make the component visible.       }
{									}
{**********************************************************************}
Begin
   Height := 28;
   Width := 28;
   Inherited Paint;
   with Canvas do
     begin
       Pen.Color := clWhite;	{ make the pen blue }
       MoveTo(0, 0);
       LineTo(Width-1, 0);
       Pen.Color := clBtnShadow;
       LineTo(Width-1, Height-1);
       LineTo(0, Height-1);
       Pen.Color := clWhite;	{ make the pen blue }
       Lineto(0, 0);
       visible := false;
     end;
End;

{--------------------Paint-------------------John Mertus---14-June--97--}

    procedure TAudioOut.Paint;

{ This LOADs the correct display.                                       }
{									}
{**********************************************************************}
Var
  Image1 :TBitMap;
Begin
   Inherited Paint;
   Image1 := TBitmap.Create;
   Image1.LoadFromResourceName(HInstance, 'TAUDIOOUT');
   Canvas.Draw(2,2,Image1);
   Image1.Free;
End;


{--------------------Paint-------------------John Mertus---14-June--97--}

    procedure TAudioIn.Paint;

{ This LOADs the correct display.                                       }
{									}
{**********************************************************************}
Var
  Image1 :TBitMap;
Begin
   Inherited Paint;
   Image1 := TBitmap.Create;
   Image1.LoadFromResourceName(HInstance, 'TAUDIOIN');
   Canvas.Draw(2,2,Image1);
   Image1.Free;
End;


End.
