{*******************************************************************************

Unit AVIFilePlayer.pas

Copyright CNRS (juin 2007)
auteur : Sylvain Mottet
sylvain.mottet @ univ-paris5.fr

Ce programme crit en Delphi permet de jouer des fichiers vidos au format
AVI sonores  l'aide d'un thread, en s'appuyant sur l'API Video For
Windows (VFW) de Microsoft.

Ce logiciel est rgi par la licence CeCILL-C soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL-C telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme, le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,  l'utilisation,  la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant des connaissances informatiques approfondies. Les
utilisateurs sont donc invits  charger et tester l'adquation du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL-C, et que vous en avez accept les
termes.

*******************************************************************************}

unit AVIFilePlayer;

interface

uses
  Classes, Controls, Graphics, Windows, VFW, Forms, MMSystem, SyncObjs,
  MMTimer2, MMWaveOut;

const
  MAX_AUDIO_STATIC_BUFFER_SIZE = 204800; // 200ko

type
  TAVIFilePlayer = class(TThread)
  private
    FCritSec: TCriticalSection;
    FFileName: String;
    FDuration: Integer;
    FPlaying: Boolean;
    FLooping: Boolean;
    FVisible: Boolean;

    AudioStream: VFW.IAVISTREAM;
    mmwo: TMMWaveOut;
    FPWavHeader: MMSystem.PWaveFormatEx;
    FAudioBufferSize: Longint;
    FSamplesToRead: Longint;
    FAudioStreamPos: Longint;
    FAudioSync: Longint;
    FSyncTime: Cardinal;

    VideoStream: VFW.IAVISTREAM;
    FFrameWidth: Integer;
    FFrameHeight: Integer;
    FGetFrame: VFW.IGETFRAME;
    FCurrentFrame: Longint;
    FLastFrame: Longint;

    procedure EndOfWave(Sender: TObject);
    procedure FillAudioBuffer(buffer: PChar; offset: Integer; var buffersize: LongWord);
    procedure SetVisible(Value: Boolean);

  protected
    procedure Execute; override;

  public
    EventOnAVIEnd: TNotifyEvent;
    HostCanvas: TCanvas;
    BgColor: TColor;
    Left,
    Top,
    Width,
    Height: Integer;
    VisibleAfterEnd: Boolean;

    debuglog: TStringList;

    property FrameWidth:  Integer read FFrameWidth;
    property FrameHeight: Integer read FFrameHeight;
    property Duration: Integer read FDuration;
    property LastFrame: Integer read FLastFrame;
    property Visible: Boolean read FVisible write SetVisible;

    constructor Create;
    function    LoadFromFile(filename: String; audio_static_buffering: Boolean = False): String;
    procedure   Play(Looping: Boolean = False);
    procedure   Stop;
    procedure   CloseFile;
    destructor  Destroy; override;
  end;

implementation

uses
dialogs,
  SysUtils;

constructor TAVIFilePlayer.Create;
begin
  inherited Create(True); // create thread suspended
  BgColor := clBlack;
  FCritSec := TCriticalSection.Create;
  debuglog := TStringList.Create;
  Priority := tpHighest;
  Visible := False;
end;

procedure TAVIFilePlayer.FillAudioBuffer(buffer: PChar; offset: Integer; var buffersize: LongWord);
var
  samplesread: Longint;
  temptime: Cardinal;
begin
  temptime := GetMilliCounter;
  AVIStreamRead(AudioStream, FAudioStreamPos, FSamplesToRead, buffer, buffersize, @buffersize, @samplesread);
//  debuglog.add('audiobuffersize: ' + IntToStr(FAudioBufferSize) + ' - size used: ' + IntToStr(buffersize));
//  debuglog.add('samplestoread: ' + IntToStr(FSamplesToRead) + ' - samples read: ' + IntToStr(samplesread));
  Inc(FAudioStreamPos, samplesread);
  if ( samplesread = FSamplesToRead ) then
  begin
    Inc(FAudioSync, samplesread * 1000 div FPWavHeader^.nSamplesPerSec);
    FSyncTime := temptime;
  end;
  debuglog.Add(IntToStr(GetMilliCounter) + ' : fillbuffer ' + IntToStr(GetMilliCounter - temptime) + 'ms');
end;

procedure TAVIFilePlayer.EndOfWave(Sender: TObject);
begin
  debuglog.Add(IntToStr(GetMilliCounter) + ' : end of wave');
  if ( FLooping = True ) then
  begin
    Play(FLooping);
  end
  else
  begin
    if ( Assigned(EventOnAviEnd) ) then
      EventOnAviEnd(Self);
    Suspended := True;
  end;
end;

function TAVIFilePlayer.LoadFromFile(filename: String; audio_static_buffering: Boolean = False): String;
var
  avifile: VFW.IAVIFile;
  lsize: Long;
  formatbuffer: PByte;
  astream, vstream: VFW.IAVISTREAM;
  psi: VFW.TAVISTREAMINFO;
begin
  Result := '';
  if ( LowerCase(FFileName) = LowerCase(filename) ) then Exit; // same file
  if ( (filename <> '') and (not FileExists(filename)) ) then
  begin
    Result := 'File <' + filename + '> does not exist.';
    Exit;
  end;

  // open avi file
  if ( AVIFileOpen(avifile, PChar(filename), OF_READ, nil) <> 0 ) then
  begin
    Result := 'Unable to open avi file <' + filename + '>.';
    Exit;
  end;

  // check presence of video stream
  if ( AVIFileGetStream(avifile, vstream, streamtypeVIDEO, 0) <> 0 ) then
  begin
    Result := 'Avi file <' + filename + '> does not contain a video stream.';
    Exit;
  end;

  // from here, file is considered opened
  CloseFile; // free previously used resources
  FFileName := filename;
  VideoStream := vstream;

  // check if audio stream exists and open it
  if ( AVIFileGetStream(avifile, astream, streamtypeAUDIO, 0) = 0 ) then
  begin
    AudioStream := astream;
    AVIStreamReadFormat(AudioStream, AVIStreamStart(AudioStream), nil, @lsize);
    GetMem(formatbuffer, lsize);
    AVIStreamReadFormat(AudioStream, AVIStreamStart(AudioStream), formatbuffer, @lsize);
    FPWAVHeader := PWaveFormatEx(formatbuffer);
    AVIStreamRead(AudioStream, 0, AVISTREAMREAD_CONVENIENT, nil, 0, @FAudioBufferSize, nil);

    // read audio format from audio stream
    mmwo := TMMWaveOut.Create;
    if ( audio_static_buffering = True ) then
    begin
      // total audio size in bytes
      lsize := AVIStreamLength(AudioStream) * FPWAVHeader^.nChannels * FPWAVHeader^.wBitsPerSample div 8;
      mmwo.PrepareBuffers(FPWAVHeader^, FillAudioBuffer, lsize, 1);
    end
    else // classic audio double buffering with "_CONVENIENT" audio buffer size
      mmwo.PrepareBuffers(FPWAVHeader^, FillAudioBuffer, FAudioBufferSize, 2);
  end;

  AVIFileRelease(avifile);

  // read video format from video stream
  AVIStreamInfo(VideoStream, psi, sizeof(psi));
//  FCodecFOURCC := FOURCCTOString(psi.fccHandler);
  FFrameWidth  := psi.rcFrame.Right - psi.rcFrame.Left;
  FFrameHeight := psi.rcFrame.Bottom - psi.rcFrame.Top;
  Self.Width  := FFrameWidth;
  Self.Height := FFrameHeight;
  FLastFrame := AVIStreamLength(VideoStream) - 1;
  FDuration  := FLastFrame * 1000 * psi.dwScale div psi.dwRate;
  FGetFrame  := AVIStreamGetFrameOpen(VideoStream, PBitmapInfoHeader(AVIGETFRAMEF_BESTDISPLAYFMT));

end;

procedure TAVIFilePlayer.Play(Looping: Boolean);
begin
  if ( HostCanvas = nil ) then Exit;
  if ( FFileName = '' ) then Exit; // not initialized

  FCritSec.Enter; // suspend thread
  FLooping := Looping;
  FCurrentFrame := -1;
  FAudioStreamPos := 0;
  debuglog.Add('Play');
  if ( mmwo <> nil ) then
  begin
    FSamplesToRead := FAudioBufferSize * 8 div FPWavHeader^.wBitsPerSample div FPWavHeader^.nChannels;
    mmwo.EventOnWavEnd := Self.EndOfWave;
    mmwo.Play;
  end;
  FAudioSync := 0;
  FSyncTime := GetMilliCounter;
  FPlaying := True;
  FCritSec.Leave;
  Suspended := False;
end;

procedure TAVIFilePlayer.Execute; // Display video frames synced with sound
var
  pbmh: Windows.PBitmapInfoHeader;
  bmi: Windows.TBitmapInfo;
  pdata: PBYTE;
  nextframe: Longint;
begin
  While ( not Terminated ) do
  begin
    nextframe := AVIStreamTimeToSample(VideoStream, FAudioSync + GetMillicounter - FSyncTime);
    debuglog.Add(IntToStr(GetMilliCounter) + ' : next frame = ' + IntToStr(nextframe) + ' / ' + IntToStr(FAudioSync + GetMillicounter - FSyncTime));
    if ( nextframe > FCurrentFrame ) then
      if ( (nextframe <= FLastFrame) and (Self.FVisible) ) then
      begin
        FCurrentFrame := nextframe;
        debuglog.Add(IntToStr(GetMilliCounter) + ' : affiche frame # ' + IntToStr(nextframe));
        pbmh := AVIStreamGetFrame(FGetFrame, FCurrentFrame);
        Assert(pbmh <> nil); // check codec if failed
        pdata := PBYTE(pbmh);
        Inc(pdata, pbmh^.biSize + pbmh^.biClrUsed * Sizeof(RGBQUAD));
        bmi.bmiHeader := pbmh^;
        StretchDIBits(HostCanvas.Handle,
                      Self.Left, Self.Top, Self.Width, Self.Height, // destination rectangle
                      0, 0, FFrameWidth, FFrameHeight, // source rectangle
                      pdata, bmi, DIB_RGB_COLORS, SRCCOPY);
      end;

      if ( nextframe > FLastFrame ) then
      begin
        if ( VisibleAfterEnd = False ) then
          Visible := False;
        if ( mmwo = nil ) then
          EndOfWave(Self);
      end;

    Sleep(2);
//    Application.ProcessMessages;
  end;
  debuglog.Add(IntToStr(GetMilliCounter) + ' : TAVIFilePlayer.Execute terminated');
end;

procedure TAVIFilePlayer.Stop;
begin
  FCritSec.Enter;
  FLooping := False;
  if ( mmwo <> nil ) then
    mmwo.Stop;
  FCritSec.Leave;
  Suspended := True;
  if ( VisibleAfterEnd = False ) then
    Visible := False;
end;

procedure TAVIFilePlayer.CloseFile;
begin
  if ( FFileName = '' ) then Exit; // no used resources
  AVIStreamRelease(VideoStream);
  if ( mmwo <> nil ) then
  begin
    AVIStreamRelease(AudioStream);
    mmwo.Free;
    mmwo := nil;
  end;
end;

procedure TAVIFilePlayer.SetVisible(Value: Boolean);
begin
  FCritSec.Enter;
  // erase AVI rectangle if Visible gets false
  if ( (FVisible = True) and (Value = False) ) then
  begin
    SelectObject(HostCanvas.Handle, GetStockObject(DC_BRUSH));
    SetDCBrushColor(HostCanvas.Handle, BgColor);
    Rectangle(HostCanvas.Handle, Self.Left, Self.Top, Self.Left + Self.Width, Self.Top + Self.Height);
  end;
  FVisible := Value;
  FCritSec.Leave;
end;

destructor TAVIFilePlayer.Destroy;
begin
  Stop;
  // allow Execute to Terminate
  Self.Visible := False;
  Terminate;
  Suspended := False;
  Wait(10);
  CloseFile;
//  debuglog.SaveToFile('AVIFilePlayer.log');
  inherited;
end;

{==============================================================================}
initialization
AVIFileInit;

{==============================================================================}
finalization
AVIFileExit;

end.
