{*******************************************************************************

Unit MCC_PMD

Copyright Sylvain Mottet (mai 2005)

sylvain.mottet@univ-paris5.fr

Ce programme informatique contient la classe de base en langage Delphi
pour accder aux priphriques PMD de Measurement Computing Corp.

Ce logiciel est rgi par la licence CeCILL soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme,  le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,   l'utilisation,   la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant  des  connaissances  informatiques approfondies.  Les
utilisateurs sont donc invits  charger  et  tester  l'adquation  du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL, et que vous en avez accept les
termes.

********************************************************************************

Historique:
-----------

30 mai 2005: premire version,
             - membres PMDHandle, DeviceCapabilities, InputReport et OutputReport
             - fonctions Create, Open, Close et Destroy
8 juin 2007: - ajout proprit <Available>

********************************************************************************

Rfrences:
-----------

- Measurement Computing Corp., Personnal Measurement Devices:
http://www.measurementcomputing.com/pmd.html
- code du driver linux sous GPL (par: Warren J. Jasper - wjasper@tx.ncsu.edu)
ftp://lx10.tx.ncsu.edu/pub/Linux/drivers/
- MSDN USB API reference:
http://msdn.microsoft.com/library/default.asp?url=/library/en-us/intinput/hh/intinput/hidfunc_83116511-237a-474a-8513-275c70b8f7fd.xml.asp
- USB HID WIN32 include file translation to delphi: HID.pas
http://www.delphi-jedi.org/APILIBRARY, ftp://delphi-jedi.org/api/hid.zip
- "The HID Page"
http://www.lvr.com/hidpage.htm

*******************************************************************************}

unit MCC_PMD;

interface

uses
  Windows,
  Hid; // ftp://delphi-jedi.org/api/hid.zip , http://www.delphi-jedi.org/APILIBRARY

type
  TPMD = class
  protected
    PMDHandle: THandle;
    DeviceCapabilities: THIDPCaps;
    InputReport: array of Byte;
    OutputReport: array of Byte;
    function    Available: Boolean;
  public
    property    IsAvailable: Boolean read Available;
    constructor Create;
    function    Open(VendorID, ProductID: Integer): String;
    procedure   Close;
    destructor  Destroy; override;
  end;

implementation

uses
  SysUtils,
  SetupApi; // ftp://delphi-jedi.org/api/hid.zip , http://www.delphi-jedi.org/APILIBRARY

constructor TPMD.Create;
begin
  inherited;
  PMDHandle := INVALID_HANDLE_VALUE;
end;

function TPMD.Open(VendorID, ProductID: Integer): String;
var
  HidGuid: TGUID;
  DeviceInfoSet: HDEVINFO; // in SetupApi
  DIData: TSPDeviceInterfaceData; // in SetupApi
  pDIDetail: PSPDeviceInterfaceDetailData; // in SetupApi
  HidHandle: THandle;
  HidAttributes: THIDDAttributes;
  PreparsedData: PHIDPPreparsedData;
  MemberIndex: Integer;
  RequiredSize, BytesReturned: Integer;
begin
  Self.Close;
  HidD_GetHidGuid(HidGuid);
  DeviceInfoSet := SetupDiGetClassDevs(@HidGuid, nil, 0, DIGCF_DEVICEINTERFACE + DIGCF_PRESENT);
  Assert (DeviceInfoSet <> HDEVINFO(INVALID_HANDLE_VALUE), 'Unable to retrieve USB HID device information.' );
  MemberIndex := 0;
  DIData.cbSize := sizeof(TSPDeviceInterfaceData);
  while ( SetupDiEnumDeviceInterfaces(DeviceInfoSet, nil, HidGuid, MemberIndex, DIData) = True ) do
  begin
    // ask for memory space required for pDIDetail.DevicePath (get RequiredSize)
    RequiredSize := 0;
    SetupDiGetDeviceInterfaceDetail(DeviceInfoSet, @DIData, nil, 0, @RequiredSize, nil);
    // allocate memory for pDIDetail.DevicePath
    pDIDetail := AllocMem(RequiredSize);
    pDIDetail.cbSize := sizeof(TSPDeviceInterfaceDetailData);
    // get pDIDetail.DevicePath
    SetupDiGetDeviceInterfaceDetail(DeviceInfoSet, @DIData, pDIDetail, RequiredSize, @BytesReturned, nil);
    // open device
    HidHandle := CreateFile(PChar(@pDIDetail.DevicePath),       // filename
                            GENERIC_READ + GENERIC_WRITE,       // access mode
                            FILE_SHARE_READ + FILE_SHARE_WRITE, // share mode
                            nil,                                // security attributes
                            OPEN_EXISTING,                      // creation disposition
                            0, 0);                              // flags and attributes, template file
    FreeMem(pDIDetail); // DevicePath not needed any more
{
//ManufacturerString, ProductString: array [0..126] of WideChar; // 126 wchar max + terminal \0
    HidD_GetManufacturerString(HidHandle, @ManufacturerString, sizeof(ManufacturerString));
    HidD_GetProductString(HidHandle, @ProductString, sizeof(ProductString));
}
    // check for VID and PID attributes
    HidAttributes.Size := sizeof(THIDDAttributes);
    HidD_GetAttributes(HidHandle, HidAttributes);
    if ( (HidAttributes.VendorID = VendorID) and (HidAttributes.ProductID = ProductID) ) then
    begin
      // device found
      PMDHandle := HidHandle;
      // get inputreport and outputreport sizes
      HidD_GetPreparsedData(HidHandle, PreparsedData);
      HidP_GetCaps(PreparsedData, DeviceCapabilities);
      SetLength(InputReport,     DeviceCapabilities.InputReportByteLength);
      fillchar( InputReport[0],  DeviceCapabilities.InputReportByteLength, 0);
      SetLength(OutputReport,    DeviceCapabilities.OutputReportByteLength);
      fillchar( OutputReport[0], DeviceCapabilities.OutputReportByteLength, 0);
      HidD_FreePreparsedData(PreparsedData);
      Break;
    end
    else
      CloseHandle(HidHandle);

    // next attached device to scan for VID and PID
    Inc(MemberIndex);
  end;

  if ( not IsAvailable ) then
    result := 'Can''t find PMD device'
  else
    result := '';
end;

function TPMD.Available: Boolean;
begin
  result := (PMDHandle <> INVALID_HANDLE_VALUE);
end;

procedure TPMD.Close;
begin
  if ( PMDHandle <> INVALID_HANDLE_VALUE ) then
    CloseHandle(PMDHandle);
  PMDHandle := INVALID_HANDLE_VALUE;
end;

destructor TPMD.Destroy;
begin
  Self.Close;
  inherited;
end;

end.
