{*******************************************************************************

Unit MCIPlayer.pas

Copyright CNRS (juin 2007)
auteur : Sylvain Mottet
sylvain.mottet @ univ-paris5.fr

Ce programme a t crit pour valuer la possibilit de jouer des fichiers
vidos au format AVI en utilisant l'API MCI de Microsoft.

Ce logiciel est rgi par la licence CeCILL-C soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL-C telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme, le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,  l'utilisation,  la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant des connaissances informatiques approfondies. Les
utilisateurs sont donc invits  charger et tester l'adquation du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL-C, et que vous en avez accept les
termes.

*******************************************************************************}

unit MCIPlayer;

interface

uses
  Forms, Windows, MMSystem;

type

  MCIErrString = String[120];

  TMCIPlayer = Class
  private
    DeviceID: MCIDeviceID;
    FDisplayRect: TRect;
    LastCmdResult: MCIERROR;

    procedure SetDisplayRect(dr: TRect);
    function  GetNumberOfFrames: Integer;
    function  GetDuration: Integer;
    function  GetLastMCIError: MCIErrString;

  public
    property DisplayRect: TRect read FDisplayRect write SetDisplayRect;
    property NumberOfFrames: Integer read GetNumberOfFrames;
    property Duration: Integer read GetDuration;
    Property LastErrorString: MCIErrString read GetLastMCIError;

    constructor Create;
    destructor  Destroy; Override;
    function    Open(filename: String): String;
    procedure   Rewind;
    function    Play(hostform: TForm; fromframe: Integer = -1; toframe: Integer = -1): String;
    procedure   Stop;
    procedure   Close;
  end;

//==============================================================================

implementation

constructor TMCIPlayer.Create;
begin
  inherited;
  DeviceID := 0;
end;

destructor TMCIPlayer.Destroy;
begin
  Self.Close;
  inherited;
end;

function TMCIPlayer.Open(filename: String): String;
var
  OpenParms: TMCI_Open_Parms;
  RectParms: TMCI_Anim_Rect_Parms;
begin
  FillChar(OpenParms, Sizeof(OpenParms), 0);
  OpenParms.lpstrElementName := PChar(filename);
  LastCmdResult := mciSendCommand(0, MCI_OPEN, MCI_WAIT or MCI_OPEN_ELEMENT, Cardinal(@OpenParms));
  Assert(LastCmdResult = 0, GetLastMCIError);
  Self.DeviceID := OpenParms.wDeviceID;
  result := GetLastMCIError;

  FillChar(RectParms, Sizeof(RectParms), 0);
  LastCmdResult := mciSendCommand(Self.DeviceID, MCI_WHERE, MCI_ANIM_WHERE_DESTINATION, Cardinal(@RectParms));
  FDisplayRect := RectParms.rc;
  Inc(FDisplayRect.Right, FDisplayRect.Left); // transfrom .width to .right
  Inc(FDisplayRect.Bottom, FDisplayRect.Top); // transfrom .height to .bottom
end;

procedure TMCIPlayer.Rewind; // marche pas.
var
  SeekParms: TMCI_Seek_Parms;
begin
  if ( Self.DeviceID = 0 ) then Exit;
  FillChar(SeekParms, Sizeof(SeekParms), 0);
  LastCmdResult := mciSendCommand(Self.DeviceID, MCI_SEEK or MCI_SEEK_TO_START, MCI_WAIT, Cardinal(@SeekParms));
end;

function TMCIPlayer.Play(hostform: TForm; fromframe: Integer = -1; toframe: Integer = -1): String;
var
  WindowParms: TMCI_Anim_Window_Parms;
  PlayParms: TMCI_Play_Parms;
begin
  if ( Self.DeviceID = 0 ) then Exit;
  FillChar(WindowParms, Sizeof(WindowParms), 0);
  WindowParms.Wnd := hostform.Handle;
  LastCmdResult := mciSendCommand(Self.DeviceID, MCI_WINDOW, MCI_WAIT or MCI_ANIM_WINDOW_HWND, Cardinal(@WindowParms));

  FillChar(PlayParms, Sizeof(PlayParms), 0);
  if ( fromframe > -1 ) then
    PlayParms.dwFrom := fromframe
  else
    PlayParms.dwFrom := 0;
  if ( toframe > -1 ) then
    PlayParms.dwTo := toframe
  else
    PlayParms.dwTo := GetNumberOfFrames;
  LastCmdResult := mciSendCommand(Self.DeviceID, MCI_PLAY, MCI_FROM or MCI_TO, Cardinal(@PlayParms));
  Assert(LastCmdResult = 0, GetLastMCIError);
end;

procedure TMCIPlayer.Stop;
var
  GenParms: TMCI_Generic_Parms;
begin
  if ( Self.DeviceID = 0 ) then Exit;
  FillChar(GenParms, Sizeof(GenParms), 0);
  LastCmdResult := mciSendCommand(Self.DeviceID, MCI_STOP, 0, Cardinal(@GenParms));
end;

procedure TMCIPlayer.Close;
var
  GenParms: TMCI_Generic_Parms;
begin
  if ( Self.DeviceID = 0 ) then Exit;
  FillChar(GenParms, Sizeof(GenParms), 0);
  LastCmdResult := mciSendCommand(Self.DeviceID, MCI_CLOSE, MCI_WAIT, Cardinal(@GenParms));
  DeviceID := 0;
end;

//==============================================================================
// properties functions

procedure TMCIPlayer.SetDisplayRect(dr: TRect);
var
  RectParms: TMCI_Anim_Rect_Parms;
begin
  if ( Self.DeviceID = 0 ) then Exit;
  FillChar(RectParms, Sizeof(RectParms), 0);
  FDisplayRect := dr;
  RectParms.rc := FDisplayRect;
  Dec(RectParms.rc.Right, RectParms.rc.Left); // transform .right to .width
  Dec(RectParms.rc.Bottom, RectParms.rc.Top); // transform .bottom to .height
  LastCmdResult := mciSendCommand(Self.DeviceID, MCI_PUT, MCI_ANIM_RECT or MCI_ANIM_PUT_DESTINATION, Cardinal(@RectParms));
end;

function TMCIPlayer.GetNumberOfFrames: Integer;
var
  StatusParms: TMCI_Status_Parms;
begin
  FillChar(StatusParms, Sizeof(StatusParms), 0);
  StatusParms.dwItem := MCI_STATUS_LENGTH;
  LastCmdResult := mciSendCommand(Self.DeviceID, MCI_STATUS, MCI_STATUS_ITEM, Cardinal(@StatusParms));
  Result := StatusParms.dwReturn;
end;

function TMCIPlayer.GetDuration: Integer;
var
  SetParms: TMCI_Set_Parms;
  StatusParms: TMCI_Status_Parms;
begin
  FillChar(SetParms, SizeOf(SetParms), 0);
  SetParms.dwTimeFormat := MCI_FORMAT_MILLISECONDS;
  LastCmdResult := mciSendCommand( Self.DeviceID, MCI_SET, MCI_SET_TIME_FORMAT, Cardinal(@SetParms) );

  FillChar(StatusParms, Sizeof(StatusParms), 0);
  StatusParms.dwItem := MCI_STATUS_LENGTH;
  LastCmdResult := mciSendCommand(Self.DeviceID, MCI_STATUS, MCI_STATUS_ITEM, Cardinal(@StatusParms));
  Result := StatusParms.dwReturn;
end;

function  TMCIPlayer.GetLastMCIError: MCIErrString;
begin
  result[0] := Char(Sizeof(MCIErrString));
  mciGetErrorString(LastCmdResult, @(result[1]), Sizeof(MCIErrString));
end;

end.
