{*******************************************************************************

Unit MMWaveOut.pas

Copyright CNRS (juin 2007)
auteur : Sylvain Mottet
sylvain.mottet @ univ-paris5.fr

Ce programme permet de jouer des sons en utilisant l'API MMSystem de Microsoft.

Ce logiciel est rgi par la licence CeCILL-C soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL-C telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme, le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,  l'utilisation,  la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant des connaissances informatiques approfondies. Les
utilisateurs sont donc invits  charger et tester l'adquation du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL-C, et que vous en avez accept les
termes.

*******************************************************************************}

unit MMWaveOut;

interface

uses
  Classes, MMSystem;

type
  TFillBufferProc = procedure(buffer: PChar; offset: Integer; var buffersize: LongWord) of object;

  TMMWaveOut = class
    hwo: HWAVEOUT;
    Wavehdrs: array of MMSystem.TWAVEHDR;
    Buffers: array of array of Byte;
    FPlaying: Boolean;
    FWaveFormatEx: MMSystem.TWaveFormatEx;
    FUserFillBuffer: TFillBufferProc;
    FBufferCount: Integer;
    FBufferSize: Integer;
    FUsedBuffers: Integer;
    FBytesRead: Integer;

    function  ReleaseHWaveOut: Boolean;

  public
    LastError: String;
    EventOnWavEnd: Classes.TNotifyEvent;

    constructor Create;
    procedure   PrepareBuffers(wave_format: TWaveFormatEx;
                               fill_buffer_proc: TFillBufferProc;
                               buffer_size: Integer = 10000;
                               buffer_number: Integer = 2);
    procedure   Play;
    procedure   Stop;
    destructor  Destroy; override;
  end;

//==============================================================================

implementation

uses
  SysUtils, Windows;

procedure WaveOutProc(hwo: HWAVEOUT; uMsg: UINT; dwUser: DWORD; dw1, dw2: DWORD) stdcall;
var
  pwh: PWAVEHDR;
begin
  if ( uMsg <> WOM_DONE ) then Exit;
  pwh := PWAVEHDR(dw1);
  with ( TMMWaveOut(dwUser) ) do
  begin
    if ( FBufferCount = 1 ) then
    begin
      // static buffering, keep all wav data in one buffer and don't refill it
      if ( Assigned(EventOnWavEnd) = True ) then
        EventOnWavEnd(TObject(dwUser));
      Exit;
    end;

    Dec(FUsedBuffers);
    if ( FPlaying = False ) then Exit; // stop requested

    pwh^.dwBufferLength := FBufferSize;
    FUserFillBuffer(pwh^.lpData, FBytesRead, pwh^.dwBufferLength);
    Inc(FBytesRead, pwh^.dwBufferLength);

    if ( pwh^.dwBufferLength > 0 ) then // play read data
    begin
      waveOutWrite(hwo, pwh, SizeOf(pwh^));
      Inc(FUsedBuffers);
    end
    else // all the data have been played
      if ( (FUsedBuffers = 0) and (Assigned(EventOnWavEnd) = True) ) then
        EventOnWavEnd(TObject(dwUser));
  end;
end;

constructor TMMWaveOut.Create;
begin
  inherited;
  FUserFillBuffer := nil;
  hwo := 0;
  FPlaying := False;
end;

function TMMWaveOut.ReleaseHWaveOut: Boolean;
var
  i: Integer;
begin
  result := True;
  if ( hwo = 0 ) then Exit;
  waveOutReset(hwo);
  for i := 0 to High(Wavehdrs) do
  begin
    waveOutUnprepareHeader(hwo, @Wavehdrs[i], SizeOf(TWAVEHDR));
    SetLength(Buffers[i], 0);
  end;
  SetLength(Buffers, 0);
  SetLength(Wavehdrs, 0);
  waveOutClose(hwo);
  hwo := 0;
end;

procedure TMMWaveOut.PrepareBuffers(wave_format: TWaveFormatEx; fill_buffer_proc: TFillBufferProc; buffer_size: Integer = 10000; buffer_number: Integer = 2);
var
  i: Integer;
  mmres: MMRESULT;
begin
  ReleaseHWaveOut; // free previous resources
  FWaveFormatEx := wave_format;
  FUserFillBuffer := fill_buffer_proc;
  FBufferCount := buffer_number;
  FBufferSize := buffer_size;

  mmres := waveOutOpen(@hwo, WAVE_MAPPER, @FWaveFormatEx, DWORD(@WaveOutProc), DWORD(Self), CALLBACK_FUNCTION);
  if ( mmres <> MMSYSERR_NOERROR ) then
  begin
    LastError := 'waveOutOpen error: ' + IntToStr(mmres);
    Exit;
  end;

  SetLength(Wavehdrs, FBufferCount);
  SetLength(Buffers, FBufferCount);
  for i := 0 to High(Wavehdrs) do
    with Wavehdrs[i] do
    begin
      dwBytesRecorded := 0;
      dwUser := 0;
      dwFlags := 0;
      dwLoops := 0;
      SetLength(Buffers[i], FBufferSize);
      lpData := PChar(@Buffers[i][0]);
      mmres := waveOutPrepareHeader(hwo, @Wavehdrs[i], SizeOf(TWAVEHDR));
      if ( mmres <> MMSYSERR_NOERROR ) then
        LastError := 'waveOutPrepareHeader error: ' + IntToStr(mmres);
    end;

  if ( FBufferCount = 1 ) then
  begin
    Wavehdrs[0].dwBufferLength := FBufferSize;
    FUserFillBuffer(Wavehdrs[0].lpData, 0, Wavehdrs[0].dwBufferLength);
  end;

end;

procedure TMMWaveOut.Play;
var
  i: Integer;
begin
  if ( hwo = 0 ) then Exit;
  FUsedBuffers := FBufferCount;
  FBytesRead := 0;
  FPlaying := True;
  if ( FBufferCount = 1 ) then
  begin
    waveOutWrite(hwo, @Wavehdrs[0], SizeOf(Wavehdrs[0]));
  end
  else
  begin
    for i := 0 to High(Wavehdrs) do
      WaveOutProc(hwo, WOM_DONE, DWORD(Self), DWORD(@Wavehdrs[i]), 0);
  end;
end;

procedure TMMWaveOut.Stop;
begin
  if ( hwo <> 0 ) then
  begin
    FPlaying := False;
    waveOutReset(hwo);
  end;
end;

destructor TMMWaveOut.Destroy;
begin
  ReleaseHWaveOut;
  inherited;
end;

end.


