{*******************************************************************************

Unit MediaFileList.pas

Copyright CNRS (juin 2007)
auteur : Sylvain Mottet
sylvain.mottet @ univ-paris5.fr

Ce programme permet de grer une liste de fichiers multimedia ainsi que
les objets pour les afficher, son principal intrt est d'viter de charger
plusieurs fois le mme fichier.

Ce logiciel est rgi par la licence CeCILL-C soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL-C telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme, le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,  l'utilisation,  la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant des connaissances informatiques approfondies. Les
utilisateurs sont donc invits  charger et tester l'adquation du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL-C, et que vous en avez accept les
termes.

*******************************************************************************}

unit MediaFileList;

interface

uses
  Classes;

const
  USEMCIPLAYER = False;

type
  TFileType = (ft_autodetect, ft_picture, ft_wav, ft_avi, ft_mci);
  TMediaFileList = class(TStringList)
    function AddFile(fname: String; ftype: TFileType = ft_autodetect): TObject;
  end;

var
  MFL: TMediaFileList;

implementation

uses
  Graphics, // TPicture
  SysUtils,
  Tools, // LoadBmpFile
  MCIPlayer,
  AVIFilePlayer,
  WAVFilePlayer;

function TMediaFileList.AddFile(fname: String; ftype: TFileType = ft_autodetect): TObject;
var
  idx: Integer;
  ext: String;
  obj: TObject;
  objpic: TPicture;
  objwav: TWAVFilePlayer;
  objavi: TAVIFilePlayer;
  objmci: TMCIPlayer;
begin
  idx := IndexOf(fname);
  if ( idx >= 0 ) then
  begin
    result := Objects[idx];
    Exit;
  end;
  ext := LowerCase(ExtractFileExt(fname));
  obj := nil;

  if ( ftype = ft_autodetect ) then
  begin
    if ( (ext = '.jpg') or (ext = '.jpeg') or (ext = '.bmp') ) then
      ftype := ft_picture;
    if ( ext = '.wav' ) then
      ftype := ft_wav;
    if ( (ext = '.mpg') or (ext = '.mpeg') ) then
      ftype := ft_mci;
    if ( ext = '.avi' ) then
      ftype := ft_avi;
  end;

  case ftype of
    ft_picture: begin
      objpic := TPicture.Create;
      LoadBmpFile(objpic.Bitmap, fname);
      obj := objpic;
    end;
    ft_wav: begin
      objwav := TWAVFilePlayer.Create;
      objwav.LoadFromFile(fname, True); // static buffering
      obj := objwav;
    end;
    ft_mci: begin
      objmci := TMCIPlayer.Create;
      objmci.Open(fname);
      obj := objmci;
    end;
    ft_avi: begin
      objavi := TAVIFilePlayer.Create;
//      objavi.Visible := False;
      objavi.LoadFromFile(fname, True);
      obj := objavi;
    end;
  end;

  if ( obj <> nil ) then
  begin
    idx := Add(fname);
    Objects[idx] := obj;
  end;
  result := obj;
end;


initialization

MFL := TMediaFileList.Create;
MFL.Sorted := True;

finalization

while ( MFL.Count > 0 ) do
begin
  MFL.Objects[0].Free;
  MFL.Delete(0);
end;

end.
