{*******************************************************************************

Unit Tools.pas

Copyright CNRS (juin 2007)
auteur : Sylvain Mottet
sylvain.mottet @ univ-paris5.fr

Ce programme fournit de multiples fonctions utilitaires.

Ce logiciel est rgi par la licence CeCILL-C soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL-C telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme, le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,  l'utilisation,  la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant des connaissances informatiques approfondies. Les
utilisateurs sont donc invits  charger et tester l'adquation du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL-C, et que vous en avez accept les
termes.

*******************************************************************************}

unit Tools;

interface

uses
  Classes, Controls, Messages, SysUtils, Windows, Graphics, Forms;

const
  fLower = Byte(1);
  fTrim  = Byte(2);

function  WordDump(w: Word): AnsiString;
function  LongWordDump (lw: Longword): AnsiString;
procedure SplitString (s, sep: String; sOut: TStringList; flags: Byte = 0);
procedure StringListShuffle (sl: TStringList);
function  strLeft(s: String; n: Integer): String;
function  strRight(s: String; n: Integer): String;
procedure kbFlush(Handle: HWND);
function  RGBStrToTColor(rgbstr: String; defaultvalue: TColor): TColor;
function  TColorToRGBStr(tcol: TColor): String;
function  LoadBmpFile(bmp: TBitmap; filename: String): Boolean;
function  Min(a, b: Cardinal): Cardinal;
function  Max(a, b: Cardinal): Cardinal;
function  CharCount(c: Char; s: String): Integer;
function  StringToNum(s: String; defaultvalue: Integer): Integer;
procedure GDIFillRect(WinCtrl: TWinControl; Left, Top, Right, Bottom: Integer; Color: TColor);

{------------------------------------------------------------------------------}

implementation

uses
  jpeg;
  
// WordDump : convertit un word en une chane de caractres ayant la
// mme reprsentation binaire (pour les E/S disque ou srie)
function WordDump(w: Word): AnsiString;
begin
  WordDump := Chr(WordRec(w).lo) + Chr(WordRec(w).hi);
end;

// LongWordDump : comme WordDump, mais sur un LongWord
function LongWordDump(lw: Longword): AnsiString;
begin
  LongWordDump := Chr(WordRec(LongRec(lw).lo).lo) + Chr(WordRec(LongRec(lw).lo).hi)
                  + Chr(WordRec(LongRec(lw).hi).lo) + Chr(WordRec(LongRec(lw).hi).hi);
end;

// SplitString : dcoupe une chane contenant des sparateurs en un
// tableau de sous-chanes
procedure SplitString(s, sep: String; sOut: TStringList; flags: Byte = 0);
var
  ideb, ifin, l: Integer;
begin
  l := Length(s);
  sOut.Clear;
  ideb := 1;
  ifin := Pos(sep, copy(s, ideb, l));
  while ifin <> 0 do
  begin
    sOut.Add(Copy(s, ideb, ifin - 1));
    ideb := ideb + ifin;
    ifin := Pos(sep, copy(s, ideb, l));
  end;
//  if ideb <> 1 then
    sOut.Add(Copy(s, ideb, l));

  for l := 0 to sOut.Count - 1 do
  begin
    if (flags and fLower) <> 0 then
      sOut[l] := LowerCase(sOut[l]);
    if (flags and fTrim) <> 0 then
      sOut[l] := Trim(sOut[l]);
  end;
end;

// StringListShuffle : mlange une liste de chanes de caractres
procedure StringListShuffle(sl: TStringList);
var
  n, i, j : Integer;
begin
  Randomize;
  for n := 0 to sl.Count * 2 do
  begin
    i := Random(sl.Count);
    j := Random(sl.Count);
    sl.Exchange(i, j);
  end;
end;

// strLeft : renvoie les n premiers caractres d'une chane
function strLeft(s: String; n: Integer): String;
begin
  strLeft := Copy(s, 1, n);
end;

// strRight : renvoie les n derniers caractres d'une chane
function strRight(s: String; n: Integer): String;
begin
  strRight := Copy(s, Length(s) - n, n);
end;

// kbFlush : vide les vnements clavier en attente pour la fentre
// dont le handle est pass en paramtre
procedure kbFlush(Handle: HWND);
var
  msg : TMSG;
begin
  while PeekMessage(msg, Handle, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE) do;
end;

// converts a string 'r,g,b' (0<=rgb<=255) to a TColor
function RGBStrToTColor(rgbstr: String; defaultvalue: TColor): TColor;
var
  i, j: Integer;
  r, g, b: Integer;
  errcode: Integer;
begin
  i := Pos(',', rgbstr);
  j := i + Pos(',', Copy(rgbstr, i + 1, Length(rgbstr)));
  if (i = 0) or (i = j) then // can't find 2 commas
  begin
    result := defaultvalue;
    Exit;
  end;
  Val(Copy(rgbstr, 1, i - 1), r, errcode);
  Val(Copy(rgbstr, i + 1, j - i - 1), g, errcode);
  Val(Copy(rgbStr, j + 1, Length(rgbStr) - j), b, errcode);
  result := r + g shl 8 + b shl 16;
end;

// converts a TColor to a string 'r,g,b' (0<=rgb<=255)
function TColorToRGBStr(tcol: TColor): String;
var
  r, g, b: Integer;
begin
  r := (tcol and 255);
  g := (tcol shr 8) and 255;
  b := (tcol shr 16) and 255;
  result := IntToStr(r) + ',' + IntToStr(g) + ',' + IntToStr(b);
end;

function LoadBmpFile(bmp: TBitmap; filename: String): Boolean;
var
  pic: TPicture;
  jpg: TJPEGImage;
  ext: String;
begin
  if ( FileExists(filename) = false ) then
  begin
    result := False;
    Exit;
  end;
  ext := LowerCase(ExtractFileExt(filename));
  if ( (ext = '.jpg') or (ext = '.jpeg') ) then
  begin
    jpg := TJPEGImage.Create;
    jpg.LoadFromFile(filename);
    with bmp do
    begin
      Width := jpg.Width;
      Height := jpg.Height;
      Canvas.Draw(0, 0, jpg);
    end;
    jpg.Free;
  end
  else
  begin
    pic := TPicture.Create;
    pic.LoadFromFile(filename);
    with bmp do
    begin
      Width := pic.Width;
      Height := pic.Height;
      Canvas.Draw(0, 0, pic.graphic);
    end;
    pic.Free;
  end;
  result := True;
end;

function Min(a, b: Cardinal): Cardinal;
begin
  if ( a <= b ) then
    result := a
  else
    result := b;
end;

function Max(a, b: Cardinal): Cardinal;
begin
  if ( a >= b ) then
    result := a
  else
    result := b;
end;

{
procedure wait(t: Longword);
var
  tini: Longword;
begin
  tini := GetTickCount;
  while ( (GetTickCount - tini) < t ) do Application.ProcessMessages;
end;
}

{
function StrTo4Char(str: String): Tfourchar;
var
  i: Integer;
begin
  result := StringOfChar(' ', 4);
  for i := 1 to min(Length(str), 4) do
    result[i - 1] := str[i];
end;
}

function  CharCount(c: Char; s: String): Integer;
var
  i: Integer;
begin
  result := 0;
  for i := 1 to Length(s) do
    if ( s[i] = c ) then result := result + 1;
end;

// renvoie le premier nombre entier trouv dans une chane de caractres
function StringToNum(s: String; defaultvalue: Integer): Integer;
var
  i, deb: Integer;
begin
  s := Trim(s);
  if ( s = '' ) then
  begin
    Result := defaultvalue;
    Exit;
  end;

  i := 1;
  while ( (i <= Length(s)) and ((s[i] < '0') or (s[i] > '9')) ) do
    Inc(i);
  deb := i;
  if ( deb > Length(s) ) then // pas de chiffre dans <s>
  begin
    Result := defaultvalue;
    Exit;
  end;

  while ( (i <= Length(s)) and (s[i] >= '0') and (s[i] <= '9') ) do
    Inc(i);

  if ( (deb > 1) and (s[deb - 1] = '-') ) then
    Dec(deb); // prend en compte les entiers ngatifs

  Result := StrToInt(Copy(s, deb, i - deb));
end;

procedure GDIFillRect(WinCtrl: TWinControl; Left, Top, Right, Bottom: Integer; Color: TColor);
var
  dc: HDC;
begin
  dc := GetDC(WinCtrl.Handle);
  SelectObject(dc, GetStockObject(DC_BRUSH));
  SetDCBrushColor(dc, Color and $00FFFFFF);
  Rectangle(dc, Left, Top, Right, Bottom);
end;

end.
