{*******************************************************************************

Unit UEvent.pas

Copyright CNRS (juin 2007)
auteur : Sylvain Mottet
sylvain.mottet @ univ-paris5.fr

Ce logiciel est rgi par la licence CeCILL-C soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL-C telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme, le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,  l'utilisation,  la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant des connaissances informatiques approfondies. Les
utilisateurs sont donc invits  charger et tester l'adquation du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL-C, et que vous en avez accept les
termes.

*******************************************************************************}

unit UEvent;

interface

uses
  Forms, Graphics, IniFiles, StdCtrls, Classes, Controls, Windows,
  AudioIo, WAVFilePlayer, MCIPlayer, AVIFilePlayer, MMTimer2;

type

  TEvent = class
  public
    Name: String;
    Filename: String;
    ParentForm: TForm;
    Duration: Integer;
    BGColor: TColor;
    ClearScreen: Boolean;
    RecordAudio: Boolean;
    WaveOutVolume: DWORD;
    Left,
    Top,
    Width,
    Height: Integer;
    Size: Integer;
    PortNumber: WORD;
    PortOutValue: WORD;
    PMD1208AOut: Integer; // value to write to PMD1208 digital IO port A, -1 if no write
    PMD1208BOut: Integer; // value to write to PMD1208 digital IO port B, -1 if no write

    AudioIn: TAudioIn;
    AudioInBuffer: array of Word;
    AudioInFilename: String;

    constructor Create(pParentForm: TForm);
    function  Load(inifile: TIniFile; section: String): Boolean; dynamic;
    procedure Start; dynamic;
    procedure Stop; dynamic;
    procedure EndOfTrial; dynamic;

    function  AudioInBufferFilled(Buffer: PChar; var Size: Integer): Boolean;
    procedure SaveSound;
  end;

  TSoundEvent = class(TEvent)
    Wave: TWAVFilePlayer;
    function  Load(inifile: TIniFile; section: String): Boolean; override;
    procedure Start; override;
    procedure Stop; override;
  end;

  TImageEvent = class(TEvent)
    Pic: TPicture;
    function  Load(inifile: TIniFile; section: String): Boolean; override;
    procedure Start; override;
    procedure Stop; override;
  end;

  TAviEvent = class(TEvent)
    Avi: TAVIFilePlayer;
    function  Load(inifile: TIniFile; section: String): Boolean; override;
    procedure Start; override;
    procedure Stop; override;
    procedure EndOfTrial; override;
  end;

  TMciEvent = class(TEvent)
    Mci: TMCIPlayer;
    function  Load(inifile: TIniFile; section: String): Boolean; override;
    procedure Start; override;
    procedure Stop; override;
    procedure EndOfTrial; override;
  end;

  TTextEvent = class(TEvent)
    TextLabel: TLabel;
    function  Load(inifile: TIniFile; section: String): Boolean; override;
    procedure Start; override;
    procedure Stop; override;
    procedure EndOfTrial; override;
  end;

  function  FindEventByName(aoevents: array of TEvent; evname: String): Integer;


//==============================================================================

implementation

uses
  SysUtils, // FileExists, etc.
  Dialogs, // MessageDlg
  MMSystem, // playsound
  Tools,
  MediaFileList,
  GWIOPM,
  UPMD1208LS;

//------------------------------------------------------------------------------
// helper functions ------------------------------------------------------------

function FindEventByName(aoevents: array of TEvent; evname: String): Integer;
var
  i: Integer;
begin
  for i := 0 to High(aoevents) do
    if ( aoevents[i].Name = evname ) then
    begin
      Result := i;
      Exit;
    end;
  Result := -1; // not found
end;

procedure SetPortW(Value: Word; Index: Word); assembler; register;
asm
  out dx, ax
end;

// TEvent ----------------------------------------------------------------------

constructor TEvent.Create(pParentForm: TForm);
begin
  inherited Create;
  ParentForm := pParentForm;
  ClearScreen := False;
  RecordAudio := False;
end;

function TEvent.Load(inifile: TIniFile; section: String): Boolean;
var
  value: String;
  ipos: Integer;
begin
  Result := True;

  // event name is section name
  Self.Name := section;

  // get event filename and check file
  Self.Filename := inifile.ReadString(section, 'filename', '');
  Self.Filename := ExpandFileName(Self.Filename);
  if ( (Self.Filename <> '') and (not FileExists(Self.Filename)) ) then
  begin
    MessageDlg(PChar('[' + section + '] : fichier <' + Self.Filename + '> non trouv.'),
               mtError, [mbOK], 0);
    Result := False; // error in event attributes
    Exit;
  end;

  // event duration
  Self.Duration := inifile.ReadInteger(section, 'duration', -2);
  if ( Self.Duration = -2 ) then
  begin
    MessageDlg(PChar('[' + section + '] : dure incorrecte.'), mtError, [mbOK], 0);
    Result := False; // error in event attributes
    Exit;
  end;

  // clearscreen and background color
  value := LowerCase(Trim(inifile.ReadString(section, 'clearscreen', '')));
  Self.ClearScreen := (value = 'yes') or (value = 'oui') or (value = '1');

  value := LowerCase(Trim(inifile.ReadString(section, 'backgroundcolor', '')));
  if ( value = '' ) then
    BGColor := -1
  else
    BGColor := RGBStrToTColor(value, -1);

  // record audio
  value := LowerCase(Trim(inifile.ReadString(section, 'recordaudio', '')));
  Self.RecordAudio := (value = 'yes') or (value = 'oui') or (value = '1');
  if( Self.RecordAudio = True ) then
  begin
    Self.AudioIn := TAudioIn.Create(nil);

    Self.AudioIn.BufferSize := 20000;
    Self.AudioIn.NumBuffers := 2;
    Self.AudioIn.FrameRate := 16000; // sample rate
    Self.AudioIn.Quantization := 16; // bits per sample
    Self.AudioIn.OnBufferFilled := Self.AudioInBufferFilled;
  end;

  // stimulus position on screen
  value := Trim(inifile.ReadString(section, 'posx', ''));
  if ( value <> '' ) then
  begin
    Self.Left := StringToNum(value, 0);
    if ( value[Length(value)] = '%' ) then
//      Self.Left := Screen.DesktopWidth * Self.Left div 100 // (50%,50%) in screen center
      Self.Left := Screen.Width * Self.Left div 100 // (50%,50%) in screen center
    else
      Self.Left := Self.Left + Screen.DesktopWidth div 2; // (0,0) in screen center
  end
  else
    Self.Left := 0;

  value := Trim(inifile.ReadString(section, 'posy', ''));
  if ( value <> '' ) then
  begin
    Self.Top := StringToNum(value, 0);
    if ( value[Length(value)] = '%' ) then
      Self.Top := Screen.Height * Self.Top div 100 // (50%,50%) in screen center
    else
      Self.Top := Self.Top + Screen.Height div 2; // (0,0) in screen center
  end
  else
    Self.Top := 0;

  value := Trim(inifile.ReadString(section, 'size', '100'));
  Self.Size := StringToNum(value, 100);

  value := Trim(inifile.ReadString(section, 'portout', ''));
  ipos := Pos(',', value);
  PortNumber := 0;
  if ( ipos > 0 ) then
  begin
    PortNumber := StrToInt(Copy(value, 1, ipos - 1));
    PortOutValue := StrToInt(Copy(value, ipos + 1, MAXINT));
  end;

  PMD1208AOut := StrToInt(inifile.ReadString(section, 'pmd1208aout', '-1'));
  PMD1208BOut := StrToInt(inifile.ReadString(section, 'pmd1208bout', '-1'));
end;

procedure TEvent.Start;
begin
  if ( ClearScreen = True ) then
    GDIFillRect(ParentForm, 0, 0, ParentForm.Width, ParentForm.Height, ParentForm.Color);

  if ( PortNumber > 0 ) then
    SetPortW(PortOutValue, PortNumber);

  if ( PMD1208LS.IsAvailable ) then
  begin
    if ( PMD1208AOut >= 0 ) then
    begin
      PMD1208LS.DIOConfig(DIO_PORTA, DIO_DIR_OUT);
      PMD1208LS.DByteOut(DIO_PORTA, PMD1208AOut);
    end;
    if ( PMD1208BOut >= 0 ) then
    begin
      PMD1208LS.DIOConfig(DIO_PORTB, DIO_DIR_OUT);
      PMD1208LS.DByteOut(DIO_PORTB, PMD1208BOut);
    end;
  end;

  if ( RecordAudio = True ) then
  begin
    Assert(WaveOutGetVolume(WAVE_MAPPER, @WaveOutVolume) = MMSYSERR_NOERROR, 'WaveOutGetVolume');
    Assert(WaveOutSetVolume(WAVE_MAPPER, 0) = MMSYSERR_NOERROR, 'WaveOutSetVolume');
    SetLength(AudioInBuffer, 0);
    Assert(AudioIn.Start = True, 'AudioIn.Start failed');
  end;
end;

procedure TEvent.Stop;
begin
  if ( RecordAudio = True ) then
  begin
    AudioIn.StopGracefully;
    SaveSound;
    WaveOutSetVolume(WAVE_MAPPER, WaveOutVolume);
  end;
end;

procedure TEvent.EndOfTrial;
begin
end;

function TEvent.AudioInBufferFilled(Buffer: pChar; var Size: Integer): Boolean;
var
  buflen: Integer;
begin
  buflen := Length(AudioInBuffer);
  SetLength(AudioInBuffer, buflen + (Size div 2));
  Move(Buffer^, AudioInBuffer[buflen], Size);
  Result := True;
end;

procedure TEvent.SaveSound;
var
  f: file;
  size, sizefmt: LongWord;
  header: TWaveFormatEx;
  msg: string;
begin
  with header do
  begin
    wFormatTag := WAVE_FORMAT_PCM;
    nChannels := 1;
    nSamplesPerSec := audioin.FrameRate;
    wBitsPerSample := audioin.Quantization;
    nBlockAlign := (wBitsPerSample div 8)*nchannels;
    nAvgBytesPerSec := nSamplesPerSec*nBlockAlign;
  end;
  size := length(AudioInBuffer)*2+36;
  sizefmt := 16;
  assignfile(f,AudioInFilename);
  rewrite(f,1);
  msg := 'RIFF';
  blockwrite(f,msg[1],4);
  blockwrite(f,size,4);
  msg := 'WAVEfmt ';
  blockwrite(f,msg[1],8);
  blockwrite(f,sizefmt,4);
  blockwrite(f,header,sizeof(TWaveFormatEx)-2);
  msg := 'data';
  blockwrite(f,msg[1],4);
  dec(size,36);
  blockwrite(f,size,4);
  blockwrite(f,AudioInBuffer[0],size);
  closefile(f);
end;

// TSoundEvent -----------------------------------------------------------------

function TSoundEvent.Load(inifile: TIniFile; section: String): Boolean;
begin
  Result := inherited Load(inifile, section);
  if ( Result = False ) then
    Exit;

  Wave := TWAVFilePlayer(MFL.AddFile(Self.Filename));
  if ( Self.Duration = -1 ) then
    Self.Duration := Wave.Duration;
end;

procedure TSoundEvent.Start;
begin
  inherited; // TEvent.Start
  Wave.Play;
end;

procedure TSoundEvent.Stop;
begin
//  PlaySound(nil, 0, SND_ASYNC);
  Wave.Stop;
  inherited;
end;

// TImageEvent -----------------------------------------------------------------

function TImageEvent.Load(inifile: TIniFile; section: String): Boolean;
begin
  Result := inherited Load(inifile, section);
  if ( Result = False ) then
    Exit;

  Pic.Free;
  Pic := TPicture(MFL.AddFile(Self.Filename));
  Self.Width  := Self.Pic.Bitmap.Width  * Self.Size div 100;
  Self.Height := Self.Pic.Bitmap.Height * Self.Size div 100;
  Self.Left := Self.Left - Self.width div 2 ;
  Self.Top  := Self.Top  - Self.Height div 2;
end;

procedure TImageEvent.Start;
begin
  inherited; // TEvent.Start
  if ( Pic <> nil ) then
    ParentForm.Canvas.StretchDraw(Rect(Self.Left, Self.Top, Self.Left + Self.Width, Self.Top + Self.Height), Pic.Graphic);
end;

procedure TImageEvent.Stop;
begin
  inherited;
end;

// TTextEvent -----------------------------------------------------------------

function TTextEvent.Load(inifile: TIniFile; section: String): Boolean;
begin
  Result := inherited Load(inifile, section);
  if ( Result = False ) then
    Exit;

  Self.TextLabel.Free;
  Self.TextLabel := TLabel.Create(nil);
  Self.TextLabel.Caption := inifile.ReadString(section, 'text', '');
  Self.TextLabel.Font.Name := 'Arial';
  Self.TextLabel.Font.Size := inifile.ReadInteger(section, 'size', 0);

  Self.TextLabel.Left := Self.Left - Self.TextLabel.Width div 2;
  Self.TextLabel.Top  := Self.Top  - Self.TextLabel.Height div 2;
  Self.TextLabel.Visible := False;
end;

procedure TTextEvent.Start;
begin
  inherited; // TEvent.Start
  if ( Self.TextLabel.Caption <> '' ) then
  begin
    Self.TextLabel.Parent := ParentForm;
    if ( Self.BGColor = -1 ) then
      Self.TextLabel.Color := ParentForm.Canvas.Brush.Color
    else
      Self.TextLabel.Color := BGColor;
    Self.TextLabel.Show;
  end;
end;

procedure TTextEvent.Stop;
begin
  inherited;
end;

procedure TTextEvent.EndOfTrial;
begin
  Self.TextLabel.Hide;
  Self.TextLabel.Parent := nil;
end;

// TAviEvent -------------------------------------------------------------------

function  TAviEvent.Load(inifile: TIniFile; section: String): Boolean;
begin
  Result := inherited Load(inifile, section);
  if ( Result = False ) then
    Exit;

  Avi.Free;
  Avi := TAVIFilePlayer(MFL.AddFile(Self.Filename, ft_avi));
  Avi.HostCanvas := ParentForm.Canvas;
  Avi.BgColor := ParentForm.Color;
  Avi.Visible := False;
  //  paramtrer dans le fichier .ini :
  Avi.VisibleAfterEnd := True;

  Self.Width  := Avi.FrameWidth  * Self.Size div 100;
  Self.Height := Avi.FrameHeight * Self.Size div 100;
  Self.Left   := Self.Left - Self.Width div 2;
  Self.Top    := Self.Top  - Self.Height div 2;

  if ( Self.Duration = -1 ) then
    Self.Duration := Avi.Duration;
end;

procedure TAviEvent.Start;
begin
  inherited;
  Avi.Left    := Self.Left;
  Avi.Top     := Self.Top;
  Avi.Width   := Self.Width;
  Avi.Height  := Self.Height;
  Avi.Visible := True;
  Avi.Play;
end;

procedure TAviEvent.Stop;
begin
  // don't stop AVI before end of trial if its duration is 0
  // (used to play simultaneous vids)
  if ( Self.Duration <> 0 ) then
  begin
    Avi.Stop;
    Avi.Visible := False;
  end;
end;

procedure TAviEvent.EndOfTrial;
begin
  Avi.Stop;
  Avi.Visible := False;
end;


// TMCIEvent -----------------------------------------------------------------

function  TMciEvent.Load(inifile: TIniFile; section: String): Boolean;
begin
  Result := inherited Load(inifile, section);
  if ( Result = False ) then
    Exit;

  Mci.Free;
  Mci := TMCIPlayer(MFL.AddFile(Self.Filename));
  Self.Width  := (Mci.DisplayRect.Right - Mci.DisplayRect.Left) * Self.Size div 100;
  Self.Height := (Mci.DisplayRect.Bottom - Mci.DisplayRect.Top) * Self.Size div 100;
  Self.Left := Self.Left - Self.Width div 2;
  Self.Top  := Self.Top  - Self.Height div 2;

  if ( Self.Duration = -1 ) then
    Self.Duration := Mci.Duration;
end;

procedure TMciEvent.Start;
begin
  inherited;
  Mci.DisplayRect := Rect(Self.Left, Self.Top, Self.Left + Self.Width, Self.Top + Self.Height);
  Mci.Play(ParentForm);
end;

procedure TMciEvent.Stop;
begin
end;

procedure TMciEvent.EndOfTrial;
begin
  Mci.Stop;
end;


//==============================================================================
{
initialization
  AudioIn := TAudioIn.Create(nil);
  AudioIn.BufferSize := 8192;
  AudioIn.NumBuffers := 4;
  AudioIn.FrameRate := 11025; // sample rate
  AudioIn.Quantization := 16; // bits per sample
  SetLength(AudioInBuffer, 0);

finalization
  AudioIn.Free;
}


//==============================================================================

initialization

try
  GWIOPM_Driver := TGWIOPM_Driver.Create;
  if ( GWIOPM_Driver.OpenSCM <> 0 ) then
    raise Exception.Create('GWIOPM: unable to open a session to the Service Control Manager.');
  GWIOPM_Driver.Install('');
  GWIOPM_Driver.Start;
  GWIOPM_Driver.DeviceOpen;
  if ( GWIOPM_Driver.LIOPM_Set_Ports($200, $3FF, True) <> ERROR_SUCCESS ) then
    raise Exception.Create('GWIOPM: can''t set ports.');
  GWIOPM_Driver.IOCTL_IOPMD_ACTIVATE_KIOPM;
except
  on E: Exception do
    MessageDlg(E.Message, mtError, [mbOk], 0);
end;

//==============================================================================

finalization

GWIOPM_Driver.IOCTL_IOPMD_DEACTIVATE_KIOPM;
GWIOPM_Driver.DeviceClose;
GWIOPM_Driver.Stop;
GWIOPM_Driver.Remove;
GWIOPM_Driver.CloseSCM;

end.
