{*******************************************************************************

Unit UFormRun.pas

Copyright CNRS (juin 2007)
auteur : Sylvain Mottet
sylvain.mottet @ univ-paris5.fr

Ce logiciel est rgi par la licence CeCILL-C soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL-C telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme, le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,  l'utilisation,  la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant des connaissances informatiques approfondies. Les
utilisateurs sont donc invits  charger et tester l'adquation du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL-C, et que vous en avez accept les
termes.

*******************************************************************************}

unit UFormRun;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  UBlock, UTrial, UEvent, MMTimer2;

type
  TFormRun = class(TForm)
    procedure FormActivate(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  public
    Name: String;
    SubjectName: String;
    SubjectID: String;
    ResultsFolder: String;
    ExpeBlocks: array of TBlock;
    ExpeTrials: array of TTrial;
    ExpeEvents: array of TEvent;
    CurrentBlock: TBlock;
    CurrentTrial: TTrial;
    CurrentEvent: TEvent;
    CurrentBlockIndex,
    CurrentTrialIndex,
    CurrentEventIndex: Integer;
    ExpeTimer: TMMTimer;
    ExpeLog: TStringList;
    RandomOrder: Boolean;
    BackgroundColor: TColor;
    StartTime: Int64;
    Running: Boolean;
    EscapeKeyPressed: Boolean;
    SpaceKeyPressed: Boolean;

    function  LoadFromFile(manipfullfilename: String): Boolean;
    procedure ResumeManip;
  end;

implementation
{$R *.DFM}

uses
  IniFiles,
  UFormProgress,
  MediaFileList,
  Tools;

function MSCountToString(mscount: Int64): String;
begin
  Result := Format('%.2d:%.2d.%.3d', [mscount div 60000, (mscount mod 60000) div 1000, mscount mod 1000]);
end;

function MSCountToFileName(mscount: Int64): String;
begin
  Result := Format('%.2d.%.2d.%.3d', [mscount div 60000, (mscount mod 60000) div 1000, mscount mod 1000]);
end;

function TFormRun.LoadFromFile(manipfullfilename: String): Boolean;
var
  PreviousDir: String;
  inif: TIniFile;
  sections: TStringList;
  sectionvalues: TStringList;
  blocknames: TStringList;
  secnum, valnum: Integer;
  svalue, evext,
  trialname,
  blockname: String;
  trialindex,
  blockindex: Integer;
  egalpos,
  optionpos,
  eventpos,
  trialpos: Integer;
  fp: TFormProgress;
begin
  Result := True;
  if ( not FileExists(manipfullfilename) ) then
  begin
    MessageDlg(PChar('Fichier <' + manipfullfilename + '> non trouv.'), mtError, [mbOK], 0);
    Result := False;
    Exit;
  end;
  GetDir(0, PreviousDir);
  ChDir(ExtractFileDir(manipfullfilename));
  Self.Name := ExtractFileName(manipfullfilename);

  inif := TIniFile.Create(manipfullfilename);
  sections := TStringList.Create;
  inif.ReadSections(sections);
//  sections.Sorted := True;

  // [general] section
  svalue := LowerCase(Trim(inif.ReadString('general', 'backgroundcolor', '')));
  Self.Color := RGBStrToTColor(svalue, clBlack);
  Self.Canvas.Brush.Color := RGBStrToTColor(svalue, clBlack);

  // list of blocks to show
  svalue := inif.ReadString('general', 'blocks', '');
  blocknames := TStringList.Create;
  SplitString(svalue, ',', blocknames, fTrim);
  for blockindex := 0 to blocknames.Count -1 do
  begin
    SetLength(ExpeBlocks, Length(ExpeBlocks) + 1);
    ExpeBlocks[blockindex] := TBlock.Create;
    ExpeBlocks[blockindex].Name := blocknames[blockindex];
  end;

  fp := TFormProgress.Create(Self);
  fp.ProgressBar.Min := 0;
  fp.ProgressBar.Max := sections.Count - 1;
  fp.ProgressBar.Position := 0;
  fp.Caption := 'Chargement en cours...';
  fp.Show;

  // read events
  for secnum := 0 to sections.Count - 1 do
    if ( (sections[secnum] <> 'general') and
         (sections[secnum] <> 'blocks') and
         (sections[secnum] <> 'trials') ) then
    begin
      fp.ProgressBar.Position := fp.ProgressBar.Position + 1;
      Application.ProcessMessages;

      sections[secnum] := Trim(sections[secnum]);
      svalue := inif.ReadString(sections[secnum], 'filename', '');
      evext := Lowercase(ExtractFileExt(svalue));

      // create the proper type event

      if ( (evext = '.jpg') or (evext = '.bmp') ) then
      begin
        // image event
        SetLength(ExpeEvents, Length(ExpeEvents) + 1);
        ExpeEvents[High(ExpeEvents)] := TImageEvent.Create(Self);
        evext := 'new event added';
      end;

      if ( evext = '.wav' ) then
      begin
        // sound event
        SetLength(ExpeEvents, Length(ExpeEvents) + 1);
        ExpeEvents[High(ExpeEvents)] := TSoundEvent.Create(Self);
        evext := 'new event added';
      end;

      if ( evext = '.avi' ) then
      begin
        svalue := LowerCase(Trim(inif.ReadString(sections[secnum], 'sound', 'yes')));
{
        if ( svalue = 'yes' ) then
        begin // use TMCIPlayer
          SetLength(ExpeEvents, Length(ExpeEvents) + 1);
          ExpeEvents[High(ExpeEvents)] := TMciEvent.Create(Self)
        end
        else
}
        begin // use TAVIFilePlayer
          SetLength(ExpeEvents, Length(ExpeEvents) + 1);
          ExpeEvents[High(ExpeEvents)] := TAviEvent.Create(Self)
        end;
        evext := 'new event added';
      end;

      if ( evext <> 'new event added' ) then
      begin
        // defaults to text event
        SetLength(ExpeEvents, Length(ExpeEvents) + 1);
        ExpeEvents[High(ExpeEvents)] := TTextEvent.Create(Self);
      end;

      // load event attributes
      if ( ExpeEvents[High(ExpeEvents)].Load(inif, sections[secnum]) = False ) then
      begin
        ExpeEvents[High(ExpeEvents)].Free;
        SetLength(ExpeEvents, Length(ExpeEvents) - 1);
      end;
    end;
  fp.Close;
  fp.Free;

  sectionvalues := TStringList.Create;

  // read trials
  sectionvalues.Sorted := False;
  inif.ReadSectionValues('trials', sectionvalues);
  sectionvalues.Sorted := True;
  for valnum := 0 to sectionvalues.Count - 1 do
  begin
    egalpos := Pos('=', sectionvalues[valnum]);
    svalue := Copy(sectionvalues[valnum], egalpos + 1, MAXINT);
    trialname := '';
    optionpos := Pos('.options', LowerCase(sectionvalues[valnum]));
    if ( optionpos > 0 ) then
      trialname := Copy(sectionvalues[valnum], 1, optionpos - 1);
    eventpos := Pos('.events', LowerCase(sectionvalues[valnum]));
    if ( eventpos > 0 ) then
      trialname := Copy(sectionvalues[valnum], 1, eventpos - 1);

    if ( trialname <> '' ) then
    begin
      trialindex := FindTrialByName(ExpeTrials, trialname);
      if ( trialindex = -1 ) then
      begin
        SetLength(ExpeTrials, Length(ExpeTrials) + 1);
        trialindex := High(ExpeTrials);
        ExpeTrials[trialindex] := TTrial.Create;
        ExpeTrials[trialindex].Name := trialname;
      end;
      if ( optionpos > 0 ) then
        ExpeTrials[trialindex].AddOptions(svalue);
      if ( eventpos > 0 ) then
        ExpeTrials[trialindex].AddEvents(svalue, ExpeEvents);
    end;
  end;

  // read blocks
  sectionvalues.Sorted := False;
  sectionvalues.Clear;
  inif.ReadSectionValues('blocks', sectionvalues);
  for valnum := 0 to sectionvalues.Count - 1 do
  begin
    egalpos := Pos('=', sectionvalues[valnum]);
    svalue := Copy(sectionvalues[valnum], egalpos + 1, MAXINT);
    blockname := '';
    optionpos := Pos('.options', LowerCase(sectionvalues[valnum]));
    if ( optionpos > 0 ) then
      blockname := Copy(sectionvalues[valnum], 1, optionpos - 1);
    trialpos := Pos('.trials', LowerCase(sectionvalues[valnum]));
    if ( trialpos > 0 ) then
      blockname := Copy(sectionvalues[valnum], 1, trialpos - 1);

    if ( blockname <> '' ) then
    begin
      blockindex := FindBlockByName(ExpeBlocks, blockname);
      if ( blockindex > -1 ) then
      begin
        if ( optionpos > 0 ) then
          ExpeBlocks[blockindex].AddOptions(svalue);
        if ( trialpos > 0 ) then
          ExpeBlocks[blockindex].AddTrials(svalue, ExpeTrials);
      end;
    end;
  end;

  for blockindex := 0 to High(ExpeBlocks) do
    ExpeBlocks[blockindex].Shuffle;

  ChDir(PreviousDir);
end;

procedure TFormRun.FormActivate(Sender: TObject);
begin
  Self.Cursor := crNone;

  ExpeTimer.Free;
  ExpeTimer := TMMTimer.Create;
  ExpeTimer.TimerMode := Oneshot;

  ExpeLog.Free;
  ExpeLog := TStringList.Create;

//  Self.WindowState := wsMaximized;
  Self.BorderStyle := bsNone;
//  Self.FormStyle := fsStayOnTop;
  Self.Left := 0;
  Self.Top := 0;
  Self.Width := Screen.DesktopWidth;
  Self.Height := Screen.DesktopHeight;

  SubjectID := SubjectName + FormatDateTime('yyyy-mm-dd-hh.nn.ss', Now);
  ResultsFolder := ExtractFilePath(Application.ExeName) + SubjectID + '\';
  CreateDir(ResultsFolder);

  CurrentBlockIndex := 0;
  CurrentTrialIndex := 0;
  CurrentEventIndex := 0;
  StartTime := 0;
  EscapeKeyPressed := False;
  SpaceKeyPressed := False;
  Running := False;

  ResumeManip;
end;

procedure TFormRun.FormClose(Sender: TObject; var Action: TCloseAction);
var
  i: Integer;
begin
  ExpeLog.SaveToFile(ResultsFolder + SubjectID + '.txt');
  for i := 0 to High(ExpeEvents) do
    ExpeEvents[i].Free;
  for i := 0 to High(ExpeTrials) do
    ExpeTrials[i].Free;
  for i := 0 to High(ExpeBlocks) do
    ExpeBlocks[i].Free;

  // ne garde pas les fichiers multimedia  
  while ( MFL.Count > 0 ) do
  begin
    MFL.Objects[0].Free;
    MFL.Delete(0);
  end;

end;


procedure TFormRun.ResumeManip;
var
  ib, it, ie: Integer;
begin
  Running := True;
  EscapeKeyPressed := False;

  for ib := CurrentBlockIndex to High(ExpeBlocks) do
  begin
    CurrentBlockIndex := ib;
    CurrentBlock := ExpeBlocks[CurrentBlockIndex];
    if ( CurrentBlock.Entered = False ) then 
    begin
      CurrentBlock.Entered := True;
      CurrentBlock.Shuffle;
      if ( CurrentBlock.PauseBefore = True ) then
      begin
        Running := False;
        EscapeKeyPressed := False;
        GDIFillRect(Self, 0, 0, Self.Width, Self.Height, Self.Color);
        Exit;
      end;
    end;
    SpaceKeyPressed := False;

    for it := CurrentTrialIndex to High(CurrentBlock.Trials) do
    begin
      CurrentTrialIndex := it;
      CurrentTrial := CurrentBlock.Trials[CurrentTrialIndex];
      if ( (CurrentTrial.PauseBefore = True) and (CurrentTrial.Entered = False) ) then
      begin
        Running := False;
        EscapeKeyPressed := False;
        CurrentTrial.Entered := True;
        Exit;
      end;
      SpaceKeyPressed := False;

      for ie := 0 to High(CurrentTrial.Events) do
      begin
        CurrentEventIndex := ie;
        CurrentEvent := CurrentTrial.Events[CurrentEventIndex];
        CurrentEvent.AudioInFilename := ResultsFolder + MSCountToFileName(GetMilliCounter - StartTime) + '-' + CurrentTrial.Name + '.wav';
        CurrentEvent.Start;
        ExpeLog.Add(MSCountToString(GetMilliCounter - StartTime) + ': ' + CurrentTrial.Name + ' - ' + CurrentEvent.Name);
        Wait(CurrentEvent.Duration);
        CurrentEvent.Stop;
        if ( EscapeKeyPressed = True ) then
        begin
          Running := False;
          EscapeKeyPressed := False;
          GDIFillRect(Self, 0, 0, Self.Width, Self.Height, Self.Color);
          Exit;
        end;
      end; // for ie := 0 to High(CurrentTrial.Events) do

      CurrentTrial.Stop;
      CurrentEventIndex := 0;
    end; // for it := CurrentTrialIndex to High(CurrentBlock.Trials) do

    CurrentBlock.Entered := False;
    CurrentTrialIndex := 0;
  end; // for ib := CurrentBlockIndex to High(ExpeBlocks) do

  PostMessage(Self.Handle, WM_CLOSE, 0, 0); // Close;
end;

procedure TFormRun.FormKeyPress(Sender: TObject; var Key: Char);
begin
  case Key of
    #27: begin
           if ( not running ) then
           begin
             PostMessage(Self.Handle, WM_CLOSE, 0, 0); // Close;
             Exit;
           end;
           EscapeKeyPressed := True;
         end;
    #32, '6', '+': begin
           if ( not Running ) then
           begin
             if ( StartTime = 0 ) then // waiting the space key to begin the experiment
               StartTime := GetMilliCounter;
             SpaceKeyPressed := True;
             ResumeManip;
           end;
         end;
    'r', 'R', '4', '-': begin
           if ( not running ) then
             if ( CurrentTrialIndex > 0 ) then
             begin
               CurrentTrial.Entered := False;
               Dec(CurrentTrialIndex); // replay trial
               ResumeManip;
             end;
         end;
    else begin
           ExpeLog.Add(MSCountToString(GetMilliCounter - StartTime) + ': ' + Key + ' - ' + IntToStr(Ord(Key)));
         end;
  end;
end;

end.
