{*******************************************************************************

Unit UPMD1208LS

Copyright Sylvain Mottet (mai 2005)

sylvain.mottet@univ-paris5.fr

Ce programme informatique crit en Delphi contient une classe drive de
la classe implmente dans le fichier MCC_PMD.pas, qui permet d'accder
aux entres / sorties numriques d'un priphrique PMD-1208LS de
Measurement Computing Corp.

Ce logiciel est rgi par la licence CeCILL soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme,  le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,   l'utilisation,   la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant  des  connaissances  informatiques approfondies.  Les
utilisateurs sont donc invits  charger  et  tester  l'adquation  du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL, et que vous en avez accept les
termes.

********************************************************************************

Historique:
-----------

30 mai 2005:  premire version,
              - constantes extraites de pmd-1208ls.h de Warren Jasper <wjasper@tx.ncsu.edu>
              - fonctions DeviceReset, BlinkLed, DIOConfig, DByteIn, DByteOut.
              - cre une variable globale d'accs au PMD1208LS
              - aprs DeviceReset, erreurs d'accs au PMD. (?)
23 juin 2006: amliorations:
              - ajout de ADCIn et ADCOut, lecture et criture sur les ports analogiques
              - ajout de AInToVolts, conversion d'entiers en entre vers volts
4 mai 2007:   - chang AInToVolts en AInToMillivolts pour avoir un rsultat de type entier
8 juin 2007:  - supprim l'Assert du Initialize pour ne plus avoir de message d'erreur quand
                le botier n'est pas branch. Le programme appelant doit tester <Available>
                pour savoir si le botier est disponible ou non.

********************************************************************************

Rfrences:
-----------

- Measurement Computing Corp., Personnal Measurement Devices:
http://www.measurementcomputing.com/pmd.html
- code du driver linux sous GPL (par: Warren J. Jasper - wjasper@tx.ncsu.edu)
ftp://lx10.tx.ncsu.edu/pub/Linux/drivers/
- MSDN USB API reference:
http://msdn.microsoft.com/library/default.asp?url=/library/en-us/intinput/hh/intinput/hidfunc_83116511-237a-474a-8513-275c70b8f7fd.xml.asp
- USB HID WIN32 include file translation to delphi: HID.pas
http://www.delphi-jedi.org/APILIBRARY, ftp://delphi-jedi.org/api/hid.zip
- "The HID Page"
http://www.lvr.com/hidpage.htm

*******************************************************************************}

unit UPMD1208LS;

interface

uses
  MCC_PMD;

const
  PMD1208LS_VID = $09DB;
  PMD1208LS_PID = $007A;

  PMD1208LS_MANUFACTURER_STRING = 'MCC';
  PMD1208LS_PRODUCT_STRING = 'PMD-1208LS';

  DIO_PORTA = $01;
  DIO_PORTB = $04;

  DIO_DIR_IN = $01;
  DIO_DIR_OUT = $00;

  MINILAB_CLOCK = 6000000; // 6 MHz clock
  LS_DELAY = 10; // originaly 20 * 1000;

  // Commands and Codes for PMD 1208-LS HID reports
  DCONFIG     = $0D;     // Configure digital port
  DIN         = $00;     // Read digital port
  DOUT        = $01;     // Write digital port
  DBIT_IN     = $02;     // Read digital port bit
  DBIT_OUT    = $03;     // Write digital port bit

  AIN         = $06;     // Read analog input channel
  AIN_SCAN    = $0E;     // Scan analog channels
  AIN_STOP    = $10;     // Stop scan
  ALOAD_QUEUE = $07;     // Load the channel/gain queue

  AOUT        = $08;     // Write analog output channel

  CINIT       = $05;     // Initialize counter
  CIN         = $04;     // Read Counter

  MEM_READ    = $09;     // Read Memory
  MEM_WRITE   = $0A;     // Write Memory

  BLINK_LED   = $0B;     // Causes LED to blink
  RESET       = $11;     // Reset USB interface
  SET_TRIGGER = $14;     // Configure external trigger
  SET_ID      = $0C;     // Set the user ID
  GET_ID      = $0F;     // Get the user ID

  OFFSET_ADJUSTMENT  = $1F00; // Offset Adjustment for the A/D        0x1F00 - 0x1F4F
  SE_GAIN_ADJUSTMENT = $1F50; // Single Ended Gain Adjustment for A/D 0x1F50 - 0x1F5F
  DE_GAIN_ADJUSTMENT = $1F60; // Differential Gain Adjustment for A/D 0x1F60 - 0x1F67
  CAL_PIN_VOLTAGE    = $1FA0; // Calibration pin voltage              0x1FA0 - 0x1FA3

  EXT_TRIG_FAILING_EDGE = 0;
  EXT_TRIG_RAISING_EDGE = 1;

  // Gain Ranges
  SE_10_00V  = $8;   // Single Ended +/- 10.0 V

  BP_20_00V  = $00;  // Differential +/- 20.0 V
  BP_10_00V  = $10;  // Differential +/- 10.0 V
  BP_5_00V   = $20;  // Differential +/- 5.00 V
  BP_4_00V   = $30;  // Differential +/- 4.00 V
  BP_2_50V   = $40;  // Differential +/- 2.50 V
  BP_2_00V   = $50;  // Differential +/- 2.00 V
  BP_1_25V   = $60;  // Differential +/- 1.25 V
  BP_1_00V   = $70;  // Differential +/- 1.00 V

  // Option values for AInScan
  AIN_EXECUTION  = $1; // 1 = single execution, 0 = continuous execution
  AIN_BURST_MODE = $2; // 1 = Burst Mode
  AIN_TRANSFER   = $4; // 1 = Block Transfer Mode
  AIN_TRIGGER    = $8; // 1 = Use External Trigger

type
  TPMD1208LS = class(TPMD)
  public
    function DeviceReset: String;
    function BlinkLed: String;
    function AInToMillivolts(ValueIn: Integer; Range: Byte): Integer;

    function DIOConfig(Port: Byte; Direction: Byte): String;
    function DByteIn(Port: Byte; var ByteIn: Byte): String;
    function DByteOut(Port: Byte; ByteOut: Byte): String;
    function ADCIn(Channel: Byte; Range: Byte; var ValueIn: Integer): String;
    function ADCOut(Channel: Byte; ValueOut: Word): String;
  end;

  PSmallint = ^Smallint;

var
  PMD1208LS: TPMD1208LS;

//==============================================================================

implementation

uses
  Windows,
  MMTimer2;

var
  errorstring: String;

//------------------------------------------------------------------------------

function TPMD1208LS.DeviceReset;
var
  WriteResult: Boolean;
  BytesWritten: DWORD;
begin
  if ( PMDHandle = INVALID_HANDLE_VALUE ) then
  begin
    result := 'PMD device not open';
    Exit;
  end;
  fillchar(OutputReport[0], DeviceCapabilities.OutputReportByteLength, 0);
  OutputReport[0] := 0; // report number
  OutputReport[1] := RESET;
  WriteResult := WriteFile(PMDHandle, OutputReport[0], DeviceCapabilities.OutputReportByteLength, BytesWritten, nil);
  Wait(LS_DELAY);
  if ( WriteResult = True ) then
    result := ''
  else
    result := 'TPMD1208LS.Reset: error writing output report.';
end;

//------------------------------------------------------------------------------

function TPMD1208LS.BlinkLed;
var
  WriteResult: Boolean;
  BytesWritten: DWORD;
begin
  if ( PMDHandle = INVALID_HANDLE_VALUE ) then
  begin
    result := 'PMD device not open';
    Exit;
  end;
  fillchar(OutputReport[0], DeviceCapabilities.OutputReportByteLength, 0);
  OutputReport[0] := 0; // report number
  OutputReport[1] := BLINK_LED; // command
  WriteResult := WriteFile(PMDHandle, OutputReport[0], DeviceCapabilities.OutputReportByteLength, BytesWritten, nil);
  Wait(LS_DELAY);
  if ( WriteResult = True ) then
    result := ''
  else
    result := 'TPMD1208LS.BlinkLed: error writing output report.';
end;

//------------------------------------------------------------------------------

function TPMD1208LS.DIOConfig(Port: Byte; Direction: Byte): String;
var
  WriteResult: Boolean;
  BytesWritten: DWORD;
begin
  if ( PMDHandle = INVALID_HANDLE_VALUE ) then
  begin
    result := 'PMD device not open';
    Exit;
  end;
  fillchar(OutputReport[0], DeviceCapabilities.OutputReportByteLength, 0);
  OutputReport[0] := 0; // report number
  OutputReport[1] := DCONFIG; // command
  OutputReport[2] := port;
  OutputReport[3] := direction;
  WriteResult := WriteFile(PMDHandle, OutputReport[0], DeviceCapabilities.OutputReportByteLength, BytesWritten, nil);
  Wait(LS_DELAY);
  if ( WriteResult = True ) then
    result := ''
  else
    result := 'TPMD1208LS.DConfig: error writing output report.';
end;

//------------------------------------------------------------------------------

function TPMD1208LS.DByteOut(Port: Byte; ByteOut: Byte): String;
var
  WriteResult: Boolean;
  BytesWritten: DWORD;
begin
  if ( PMDHandle = INVALID_HANDLE_VALUE ) then
  begin
    result := 'PMD device not open';
    Exit;
  end;
  fillchar(OutputReport[0], DeviceCapabilities.OutputReportByteLength, 0);
  OutputReport[0] := 0; // report number
  OutputReport[1] := DOUT; // command
  OutputReport[2] := port;
  OutputReport[3] := ByteOut;
  WriteResult := WriteFile(PMDHandle, OutputReport[0], DeviceCapabilities.OutputReportByteLength, BytesWritten, nil);
  Wait(LS_DELAY);
  if ( WriteResult = True ) then
    result := ''
  else
    result := 'TPMD1208LS.DByteOut: error writing output report.';
end;

//------------------------------------------------------------------------------

function TPMD1208LS.DByteIn(Port: Byte; var ByteIn: Byte): String;
var
  WriteResult: Boolean;
  ReadResult: Boolean;
  BytesWritten: DWORD;
  BytesRead: DWORD;
begin
  if ( PMDHandle = INVALID_HANDLE_VALUE ) then
  begin
    result := 'PMD device not open';
    Exit;
  end;
  fillchar(OutputReport[0], DeviceCapabilities.OutputReportByteLength, 0);
  OutputReport[0] := 0; // report number
  OutputReport[1] := DIN; // command
  OutputReport[2] := port;
  WriteResult := WriteFile(PMDHandle, OutputReport[0], DeviceCapabilities.OutputReportByteLength, BytesWritten, nil);
  Wait(LS_DELAY);
  if ( WriteResult = False ) then
  begin
    result := 'TPMD1208LS.DByteIn: error writing output report.';
    Exit;
  end;
  fillchar(InputReport[0], DeviceCapabilities.OutputReportByteLength, 0);
  ReadResult := ReadFile(PMDHandle, InputReport[0], DeviceCapabilities.InputReportByteLength, BytesRead, nil);
  Wait(LS_DELAY);
  if ( ReadResult = True ) then
  begin
    result := '';
    ByteIn := InputReport[1];
  end
  else
    result := 'TPMD1208LS.DByteIn: error reading input report.';
end;

//------------------------------------------------------------------------------

function TPMD1208LS.ADCIn(Channel: Byte; Range: Byte; var ValueIn: Integer): String;
var
  WriteResult: Boolean;
  ReadResult: Boolean;
  BytesWritten: DWORD;
  BytesRead: DWORD;
begin
  if ( PMDHandle = INVALID_HANDLE_VALUE ) then
  begin
    result := 'PMD device not open';
    Exit;
  end;

  // test Channel value
  if ( Range = SE_10_00V ) then
    if ( Channel > 7 ) then
    begin
      result := 'TPMD1208LS.ADCIn: channel out of range for single-ended mode.';
      Exit;
    end
    else
      if ( Channel > 3 ) then
      begin
        result := 'TPMD1208LS.ADCIn: channel out of range for differential mode.';
        Exit;
      end;

  OutputReport[0] := 0; // report number
  OutputReport[1] := AIN; // command
  OutputReport[2] := Channel;
  OutputReport[3] := Range;
  WriteResult := WriteFile(PMDHandle, OutputReport[0], DeviceCapabilities.OutputReportByteLength, BytesWritten, nil);
  Wait(LS_DELAY);
  if ( WriteResult = False ) then
  begin
    result := 'TPMD1208LS.ADCIn: error writing output report.';
    Exit;
  end;

  fillchar(InputReport[0], DeviceCapabilities.OutputReportByteLength, 0);
  ReadResult := ReadFile(PMDHandle, InputReport[0], DeviceCapabilities.InputReportByteLength, BytesRead, nil);
  Wait(LS_DELAY);
  if ( ReadResult = False ) then
  begin
    result := 'TPMD1208LS.ADCIn: error reading input report.';
    Exit;
  end;

  ValueIn := Integer(InputReport[2]) shl 8 + InputReport[1];

  if ( Range = SE_10_00V ) then
    ValueIn := ValueIn div 32 - 512
  else
    ValueIn := ValueIn div 16 - 1024; // ??

end;

//------------------------------------------------------------------------------

function TPMD1208LS.ADCOut(Channel: Byte; ValueOut: Word): String;
var
  WriteResult: Boolean;
  BytesWritten: DWORD;
begin
  if ( PMDHandle = INVALID_HANDLE_VALUE ) then
  begin
    result := 'PMD device not open';
    Exit;
  end;
  fillchar(OutputReport[0], DeviceCapabilities.OutputReportByteLength, 0);
  OutputReport[0] := 0; // report number
  OutputReport[1] := AOUT; // command
  OutputReport[2] := Channel;
  OutputReport[3] := Byte(ValueOut and $FF);
  OutputReport[3] := Byte((ValueOut shr 8) and $FF);
  WriteResult := WriteFile(PMDHandle, OutputReport[0], DeviceCapabilities.OutputReportByteLength, BytesWritten, nil);
  Wait(LS_DELAY);
  if ( WriteResult = True ) then
    result := ''
  else
    result := 'TPMD1208LS.ADCOut: error writing output report.';
end;

function TPMD1208LS.AInToMillivolts(ValueIn: Integer; Range: Byte): Integer;
begin
  if ( Range = SE_10_00V ) then
    Result := ValueIn * 1000 div 64
  else
    Result := ValueIn * 1000 div 32;
{
  case Range of
    SE_10_00V: Result := ValueIn / 51.2;
    BP_20_00V: Result := ValueIn;
    BP_10_00V: Result := ValueIn;
    BP_5_00V : Result := ValueIn;
    BP_4_00V : Result := ValueIn;
    BP_2_50V : Result := ValueIn;
    BP_2_00V : Result := ValueIn;
    BP_1_25V : Result := ValueIn;
    BP_1_00V : Result := ValueIn;
  else
    Result := 0.0;
  end;
}
end;

//==============================================================================

initialization
  PMD1208LS := TPMD1208LS.Create;
  errorstring := PMD1208LS.Open(PMD1208LS_VID, PMD1208LS_PID);

//==============================================================================

finalization
  PMD1208LS.Free;

end.
