{*******************************************************************************

Unit WAVFilePlayer.pas

Copyright CNRS (juin 2007)
auteur : Sylvain Mottet
sylvain.mottet @ univ-paris5.fr

Ce programme permet de jouer des fichiers sons au format WAV en appelant l'unit
MMWaveOut.

Ce logiciel est rgi par la licence CeCILL-C soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL-C telle que diffuse par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilit au code source et des droits de copie,
de modification et de redistribution accords par cette licence, il n'est
offert aux utilisateurs qu'une garantie limite.  Pour les mmes raisons,
seule une responsabilit restreinte pse sur l'auteur du programme, le
titulaire des droits patrimoniaux et les concdants successifs.

A cet gard  l'attention de l'utilisateur est attire sur les risques
associs au chargement,  l'utilisation,  la modification et/ou au
dveloppement et  la reproduction du logiciel par l'utilisateur tant
donn sa spcificit de logiciel libre, qui peut le rendre complexe 
manipuler et qui le rserve donc  des dveloppeurs et des professionnels
avertis possdant des connaissances informatiques approfondies. Les
utilisateurs sont donc invits  charger et tester l'adquation du
logiciel  leurs besoins dans des conditions permettant d'assurer la
scurit de leurs systmes et ou de leurs donnes et, plus gnralement,
 l'utiliser et l'exploiter dans les mmes conditions de scurit.

Le fait que vous puissiez accder  cet en-tte signifie que vous avez
pris connaissance de la licence CeCILL-C, et que vous en avez accept les
termes.

*******************************************************************************}

unit WAVFilePlayer;

interface

uses
  Classes, MMSystem, MMWaveOut;

const
  WAVE_FORMAT_ADPCM = 2; // missing from MMSystem.pas, found in mmreg.h

type
  TWAVFilePlayer = class
    FFileName: String;
    FWaveFileHandle: File;
    FWavDataPos: Integer;
    FWavDataLength: Integer;
    FWavHeader: MMSystem.TWaveFormatEx;
    mmwo: TMMWaveOut;

    FNumberOfSamples:  Integer;
    FDuration:         Integer;

    procedure FillBuffer(buffer: PChar; offset: Integer; var buffersize: LongWord);

  public
    LastError: String;
    EventOnWavEnd: Classes.TNotifyEvent;

    property    BitsPerSample: Word read FWavHeader.wBitsPerSample;
    property    Channels: Word read FWavHeader.nChannels;
    property    SamplesPerSecond: Cardinal read FWavHeader.nSamplesPerSec;
    property    NumberOfSamples: Integer read FNumberOfSamples;
    property    Duration: Integer read FDuration;

    constructor Create;
    function    LoadFromFile(filename: String; audio_static_buffering: Boolean = False): Boolean;
    procedure   Play;
    procedure   Stop;
    destructor  Destroy; override;
  end;

implementation

uses
  SysUtils, Windows, MMTimer2;

constructor TWAVFilePlayer.Create;
begin
  inherited;
  FillMemory(@FWavHeader, Sizeof(FWavHeader), 0); 
end;

procedure TWAVFilePlayer.FillBuffer(buffer: PChar; offset: Integer; var buffersize: LongWord);
begin
  Seek(FWaveFileHandle, FWavDataPos + offset);
  if ( offset + buffersize > FWavDataLength ) then
    buffersize := FWavDataLength - offset;
  BlockRead(FWaveFileHandle, buffer^, buffersize);
end;


function TWAVFilePlayer.LoadFromFile(filename: String; audio_static_buffering: Boolean = False): Boolean;
var
  strbuff: ShortString;
  buffer: array of Byte;
  filelength, chunklength: DWORD;
begin
  result := True;
  if ( FileExists(Trim(filename)) = False ) then
  begin
    result := False;
    Exit;
  end;

  // open wav file in read only mode
  FFileName := Trim(filename);
  FileMode := fmOpenRead; // read-only
  AssignFile(FWaveFileHandle, FFileName);
  Reset(FWaveFileHandle, 1);

  // RIFF chunk
  strbuff := StringOfChar(' ', 4);
  BlockRead(FWaveFileHandle, strbuff[1], 4);
  if ( strbuff <> 'RIFF' ) then
  begin
    LastError := 'File <' + filename + '> is not a RIFF file.';
    FFileName := '';
    result := False;
    Exit;
  end;
  BlockRead(FWaveFileHandle, filelength, 4);

  // WAVE identifier
  strbuff := StringOfChar(' ', 4);
  BlockRead(FWaveFileHandle, strbuff[1], 4);
  if ( strbuff <> 'WAVE' ) then
  begin
    LastError := 'File <' + filename + '> is not a WAVE file.';
    FFileName := '';
    result := False;
    Exit;
  end;

  // browse chunks
  while ( Eof(FWaveFileHandle) = False ) do
  begin
    strbuff := StringOfChar(' ', 4);
    BlockRead(FWaveFileHandle, strbuff[1], 4);
    BlockRead(FWaveFileHandle, chunklength, 4);
    // format chunk
    if ( strbuff = 'fmt ' ) then
    begin
      BlockRead(FWaveFileHandle, FWavHeader, SizeOf(TPCMWaveFormat));
      if ( (FWavHeader.wFormatTag <> WAVE_FORMAT_PCM)
           and (FWavHeader.wFormatTag <> WAVE_FORMAT_ADPCM) ) then
      begin
        LastError := 'File <' + filename + '> is not a PCM WAVE file.';
        result := False;
        Break;
      end;
      if ( chunklength > SizeOf(TPCMWaveFormat) ) then
      begin
        chunklength := chunklength - SizeOf(TPCMWaveFormat);
        SetLength(buffer, chunklength);
        BlockRead(FWaveFileHandle, buffer[0], chunklength);
      end;
      continue;
    end;
    // data chunk
    if ( strbuff = 'data' ) then
    begin
      FWavDataLength := chunklength;
      FWavDataPos := FilePos(FWaveFileHandle);
      Break;
    end;
    // drop any other chunk
    SetLength(buffer, chunklength);
    BlockRead(FWaveFileHandle, buffer[0], chunklength);
  end;

  FNumberOfSamples := FWavDataLength div FWavHeader.nBlockAlign;
  FDuration        := FNumberOfSamples * 1000 div FWavHeader.nSamplesPerSec;

  mmwo.Free;
  mmwo := TMMWaveOut.Create;
  if ( audio_static_buffering = True ) then
    mmwo.PrepareBuffers(FWavHeader, FillBuffer, FWavDataLength, 1)
  else
    mmwo.PrepareBuffers(FWavHeader, FillBuffer);
end;

procedure TWAVFilePlayer.Play;
begin
  if ( mmwo = nil ) then Exit;
  mmwo.EventOnWavEnd := Self.EventOnWavEnd;
  mmwo.Play;
end;

procedure TWAVFilePlayer.Stop;
begin
  if ( mmwo = nil ) then Exit;
  mmwo.Stop;
end;

destructor TWAVFilePlayer.Destroy;
begin
  if ( mmwo <> nil ) then
  begin
    mmwo.Stop;
    mmwo.Free;
  end;
  inherited;
end;

end.
